import glob
import os
import tempfile
from datetime import datetime

import h5py
import numpy as np
import pandas as pd
import pkg_resources
import tifffile
from hdmf.backends.hdf5 import H5DataIO
from nptdms import TdmsFile
from pynwb import get_class, load_namespaces, NWBFile, NWBHDF5IO, TimeSeries
from pynwb.file import Subject
from pynwb.image import ImageSeries
from pynwb.ophys import ImageSegmentation, OpticalChannel, TwoPhotonSeries
from pytz import timezone

from . import metadata
from .header import LabViewHeader, LabViewVersions
from .imaging import Modes
from .timings import LabViewTimings231, LabViewTimingsPre2018

try:
    import av
except ImportError:
    # This dependency is optional
    av = None


class NwbFile():
    """Silver Lab wrapper for the NWB data format.data

    Compression of data is enabled by default, but can be set to False by the user.

    Designed to be used as a context manager, i.e. do something like:
    >>> with NwbFile(output_file_path) as nwb:
    ...     nwb.import_labview_folder(folder_path, compress=True)

    However, there is also an explicit close() method, and this will be called
    when the object is deleted.

    Once a file has been opened, two main access mechanisms are provided:
    - nwb.nwb_file - the NWB API interface to the file
    - nwb.hdf_file - the h5py interface to the file

    You can also access nodes within the file using dictionary-style access, e.g.
    `nwb['/path/to/node']`.
    """

    SILVERLAB_NWB_VERSION = '0.2'

    def __init__(self, nwb_path, mode='r', verbose=True):
        """Create an interface to an NWB file.

        :param nwb_path: the NWB file to access
        :param mode: mode of file access. As for the NWB API, must be one of:
            'r'  - Readonly, file must exist
            'r+' - Read/write, file must exist
            'w'  - Create file, replacing if exists
            'w-' - Create file, fail if exists
            'a'  - Read/write if exists, create otherwise
        :param verbose: if True, print status information as processing happens
        """
        self.verbose = verbose
        self.nwb_file = None
        self.nwb_path = nwb_path
        assert mode in {'r', 'r+', 'w', 'w-', 'a'}
        self.nwb_open_mode = mode
        if mode in {'r', 'r+'} or (mode == 'a' and os.path.isfile(nwb_path)):
            self.open_nwb_file()
        # assume silverlab extension is in this file's directory
        load_namespaces(pkg_resources.resource_filename(__name__, "silverlab.namespace.yaml"))
        self.custom_silverlab_dict = dict()
        self.labview_version = None
        self.imaging_info = None
        self.trial_times = None
        self.compress = None

    def import_labview_folder(self, folder_path, compress=True):
        """Import all data from a Labview export folder into this NWB file.

        This calls three helper methods to do most of the work, to support unit tests
        of just parts of the import code.

        This will automatically import video data only if the video folder is adjacent to
        the main labview folder, with the same name but having ' VidRec' appended. If you
        use a different layout you will need to call read_video_data() separately.

        :param folder_path: the folder to import
        """
        assert os.path.isdir(folder_path)
        folder_name = os.path.basename(folder_path)
        session_id = folder_name.split(' ')[0]  # Drop the ' FunctAcq' part
        self.log('Importing Labview session', session_id, 'from', folder_path)
        self.compress = compress
        speed_data, expt_start_time = self.create_nwb_file(folder_path, session_id)
        self.add_core_metadata()
        self.import_labview_data(folder_path, folder_name, speed_data, expt_start_time)
        self.log('All data imported')

    def create_from_metadata(self, metadata_file, user=None):
        """Create a base NWB file containing only experimental metadata."""
        assert self.nwb_open_mode == 'w', ("Can only create a file if opening "
                                           "in 'w' mode")
        if not os.path.isfile(metadata_file):
            raise ValueError(f"Could not find file {metadata_file}.")
        # Read metadata file to retrieve information required by NWB
        # This includes the description and start time for the session
        self.user_metadata = metadata.read_custom_config(metadata_file)
        sessions = self.user_metadata.get('sessions')
        if not sessions:
            raise ValueError("No sessions found in file!")
        if not user:
            # Select the user if it's not ambiguous, otherwise fail
            if len(sessions) > 1:
                raise ValueError('Multiple users found in file.')
            user = list(sessions.keys())[0]
        self.record_metadata(user)
        # For now, assume that the session start time is recorded in the file
        # (normally we would get this from the LabView data).
        try:
            start_time = sessions[user]['start_time']
        except KeyError:
            raise ValueError("Start time for session not found!")
        start_time = pd.to_datetime(
            start_time, infer_datetime_format=True).tz_localize(
            timezone('Europe/London'))
        nwb_settings = {
            'session_start_time': start_time.to_pydatetime(),
            'identifier': f"{os.path.basename(metadata_file)}; {datetime.now()}",
            'session_description': self.session_description,
        }
        self.nwb_file = NWBFile(**nwb_settings)
        self.add_core_metadata()
        self.add_custom_silverlab_data(include_opto=False)
        self.log('All metadata added')

    @property
    def hdf_file(self):
        """Access the h5py interface to this NWB file."""
        assert self.nwb_file is not None
        # TODO Should we check if the file is still open? Or perhaps
        # turn this into a real field?
        return h5py.File(self.nwb_path, self.nwb_open_mode)

    def __getitem__(self, name):
        """Provide access to nodes within this file just like h5py does.

        :param name: full path (within the file) of the HDF5 node to access
        """
        return self.hdf_file[name]

    def open_nwb_file(self):
        """Open an existing NWB file for reading and optionally modification.

        TODO: If allowing modification then the copy_append setting defaults to True and
        we can't modify it via nwb_file.open - we'd need to call the underlying routines
        directly if we want to avoid copying the original file! However, this copying
        behaviour does guard against data corruption, so might well be desirable.
        """
        self.log("Opening file {}", self.nwb_path)
        with NWBHDF5IO(self.nwb_path, 'r') as io:
            self.nwb_file = io.read()

    def create_nwb_file(self, folder_path, session_id):
        """Create a new NWB file and add general lab/session metadata.

        :param folder_path: the Labview folder to import
        :param session_id: the unique session ID for this experiment
        :returns: (speed_data, expt_start_time) for passing to import_labview_data
        """

        def rel(file_name):
            """Return the path of a file name relative to the Labview folder."""
            return os.path.join(folder_path, file_name)

        # Check we're allowed to create a new file
        if not (self.nwb_open_mode == 'w' or (self.nwb_open_mode in {'a', 'w-'} and
                                              not os.path.isfile(self.nwb_path))):
            raise ValueError('Not allowed to create/overwrite {} in mode {}'.format(
                self.nwb_path, self.nwb_open_mode))
        # Figure out the metadata required when creating a new NWB file
        self.read_user_config()
        header_fields = self.parse_experiment_header_ini(rel('Experiment Header.ini'))
        speed_data, expt_start_time = self.read_speed_data(rel('Speed_Data/Speed data 001.txt'))
        localized_start_time = expt_start_time.tz_localize(timezone('Europe/London'))
        # Create the NWB file
        extensions = ["e-labview.py", "e-pixeltimes.py"]
        for i, ext in enumerate(extensions):
            extensions[i] = pkg_resources.resource_filename(__name__, ext)
        nwb_settings = {
            'session_start_time': localized_start_time.to_pydatetime(),
            'identifier': "{}; {}".format(session_id, datetime.now()),
            'session_description': self.session_description,
            'session_id': session_id,
        }
        self.nwb_file = NWBFile(**nwb_settings)
        # TODO Incorporate extensions according to new API
        self.add_labview_header(header_fields)
        # Write the new NWB file
        self._write()
        return speed_data, expt_start_time

    def add_core_metadata(self):
        """Add core metadata from the YAML config file to the NWB file.

        This fills in many of the fields in /general.
        """
        self.add_general_info('experimenter', self.user['name'])  # TODO: Add ORCID etc.
        self.add_general_info('experiment_description', self.experiment['description'])
        self.add_general_info('institution', 'University College London')
        self.add_general_info('lab', 'Silver Lab (http://silverlab.org)')
        self.add_devices_info()
        for field in ['data_collection', 'pharmacology', 'protocol', 'slices',
                      'surgery', 'virus', 'related_publications', 'notes']:
            if field in self.experiment:
                self.add_general_info(field, self.experiment[field])
        # Stimulus information is now accessed as `stimulus_notes` in the API,
        # even though it is still stored under /general/stimulus
        if 'stimulus' in self.experiment:
            self.add_general_info('stimulus_notes', self.experiment['stimulus'])
        if 'subject' in self.experiment:
            self.add_subject(self.experiment['subject'])
        # Update the file on disk:
        self._write()

    def add_subject(self, subject_data):
        """Add the valid subject information from YAML config to the NWB file.

        Use pynwb default if a subject information entry is None.
        """
        valid_subject_data = {key: value for key, value in subject_data.items() if value is not None}
        subject = Subject(**valid_subject_data)
        self.add_general_info('subject', subject)

    def import_labview_data(self, folder_path, folder_name, speed_data, expt_start_time):
        """Import the bulk of the Labview data to NWB.

        :param folder_path: the Labview folder to import
        :param folder_name: the name of the Labview folder
        :param speed_data: mouse speed data
        :param expt_start_time: when the experiment started
        """

        def rel(file_name):
            """Return the path of a file name relative to the Labview folder."""
            return os.path.join(folder_path, file_name)

        self.add_speed_data(speed_data, expt_start_time)
        self.determine_trial_times()
        self.add_stimulus()
        self.read_cycle_relative_times(folder_path)
        self.read_zplane(rel('Zplane_Pockels_Values.dat'))
        self.read_zstack(rel('Zstack Images'))
        self.add_rois(rel('ROI.dat'))
        self.read_functional_data(rel('Functional imaging TDMS data files'))
        video_folder = os.path.join(os.path.dirname(folder_path), folder_name + ' VidRec')
        if os.path.isdir(video_folder):
            self.read_video_data(video_folder)

    def __enter__(self):
        """Return this object itself as a context manager."""
        return self

    def __exit__(self, type, value, traceback):
        """Close the NWB file when the context is exited."""
        self.close()

    def __del__(self):
        """Close the NWB file when this object is destroyed."""
        self.close()

    def close(self):
        """Close our NWB file. Note that any unwritten changes will be lost."""
        self.nwb_file = None

    def log(self, msg_template, *args, **kwargs):
        """Log status information if in verbose mode.

        :param msg_template: message to log, optionally with {} placeholders
        :param args: positional arguments to pass to msg_template.format
        :param kwargs: keyword arguments to pass to msg_template.format
        """
        if self.verbose:
            import time
            timestamp = time.strftime('%H:%M:%S ')
            print(timestamp + msg_template.format(*args, **kwargs))

    def read_user_config(self):
        """Read the user configuration YAML files.

        We first read the default configuration supplied with this package.
        Then we look in the user_config_dir (as defined by appdirs) for any
        machine- & user-specific configuration settings, which override the
        defaults.
        """
        if os.path.isfile(metadata.user_conf_path):
            self.log('Reading user metadata from {}', metadata.user_conf_path)
        self.user_metadata = metadata.read_user_config()[0]  # second item in tuple are comments, only used in GUI
        return self.user_metadata

    def add_general_info(self, label, value):
        """Add a general piece of information about the experiment if it is specified.

        We allow optional values to not be specified, in which case they will be None,
        and pynwb defaults will be used.
        """
        # In the new API version, most labels are now attributes of the NWB file
        # itself (including experiment_description, lab, and others). However,
        # it appears that the file object also supports assignment of arbitrary
        # attributes, so we do not need to check the label here. The only case
        # that causes a problem is when we set an attribute that is already set,
        # which raises an AttributeError. The error message there is clear
        # enough that I don't think we need to raise a more specific exception.
        if value is not None:
            setattr(self.nwb_file, label, value)

    def add_devices_info(self):
        """Populate /general/devices with information about the rig.

        The names and descriptions of devices are taken from the metadata config file.
        """
        for device_name, desc in self.user_metadata['devices'].items():
            if not device_name.endswith('Cam'):
                # Calling create_device immediately adds the new device to the
                # NWB file
                # HACK We're adding the description where the source should go,
                # since the "help" attribute (which seems a better fit for it)
                # will not retain any overwritten value in the written-out file.
                # Oddly, this still works even after the "source" attribute was
                # removed from the API (although the description is not written
                # out to file).
                self.nwb_file.create_device(device_name, desc)

    def parse_experiment_header_ini(self, filename):
        """Read the LabView .ini file and store fields for later processing.

        The information will be stored in self.labview_header as a nested dict
        for easy access of key fields later. Those fields that are numbers will be
        stored as floating point values; everything else will be strings. Values given
        in double quotes will have the quotes removed.

        This method also uses the header info to set self.mode, which stores the
        type of imaging being performed, and figure out which user's metadata to load.

        :param filename: path to the Labview header
        :returns: the raw Labview fields as a list of lists of strings
        """
        self.log('Parsing Labview header {}', filename)
        header = LabViewHeader.from_file(filename)
        self.labview_version = header.version
        self.mode = header.imaging_mode
        # Store the imaging-related information on the file. This duplicates parts of
        # the header, but allows us to avoid extracting it repeatedly.
        self.imaging_info = header.get_imaging_information()
        # TODO this should probably go into a determine_trial_times function
        #  that hides the handling of the different LabView versions.
        if self.labview_version is LabViewVersions.v231:
            self.determine_trial_times_from_header(header)
        # Use the user specified in the header to select default session etc. metadata
        self.record_metadata(header['LOGIN']['User'])
        return header.get_raw_fields()

    def record_metadata(self, user):
        """Record information about the user and experiment on this object.

        Sets some internal fields to be used downstream. The user metadata
        configuration must have been read first.
        """
        assert self.user_metadata
        if user not in self.user_metadata['sessions']:
            if 'last_session' in self.user_metadata:
                self.log("Labview user '{}' not found in metadata;"
                         " using last session by '{}' instead.",
                         user, self.user_metadata['last_session'])
                user = self.user_metadata['last_session']
            else:
                raise ValueError("No session information found for user '{}' - please edit the"
                                 " metadata.yaml file to include their details.".format(user))
        if user not in self.user_metadata['people']:
            raise ValueError("No information found for user '{}' - please edit the metadata.yaml"
                             " file to include their details.".format(user))
        self.user = self.user_metadata['people'][user]
        expt = self.user_metadata['sessions'][user]['experiment']
        if expt not in self.user_metadata['experiments']:
            raise ValueError("Experiment '{}' not found in metadata.yaml.".format(expt))
        self.experiment = self.user_metadata['experiments'][expt]
        self.session_description = self.user_metadata['sessions'][user]['description']

    def determine_trial_times_from_header(self, header):
        self.trial_times = header.determine_trial_times()

    def add_labview_header(self, fields):
        """Add the Labview header fields verbatim to the NWB file.

        We use a fixed length ASCII string array, null-padded to the length of the longest
        string, at /general/labview_header. This is defined by one of our NWB extensions.
        It is likely to be the most portable representation for this kind of data.

        :param fields: the raw Labview headers as a list of 3-element lists of strings
        """
        self.add_general_info("labview_header", fields)  # TODO use the extension

    def add_time_series_data(self, label, data, times, ts_attrs={}, data_attrs={},
                             kind=TimeSeries):
        """Create a basic acquisition timeseries and add to the NWB file.

        :param label: Name of the group within /acquisition/timeseries.
        :param data: The data array.
        :param times: The timestamps array.
        :param ts_attrs: Any attributes for the timeseries group itself.
        :param data_attrs: Any attributes for the data array.
        :param kind: The class of timeseries to create, e.g. TwoPhotonSeries.
        :param compress: True if data should be compressed, False otherwise
        :returns: The new timeseries group.
        """
        all_attrs = dict(ts_attrs)
        all_attrs.update(data_attrs)
        if self.compress and data is not None:
            wrapped_data = H5DataIO(data=data,
                                    compression='gzip',
                                    compression_opts=4,
                                    )
        else:
            wrapped_data = data
        ts = kind(name=label, data=wrapped_data, timestamps=times, **all_attrs)
        return self.nwb_file.add_acquisition(ts)

    def read_speed_data(self, file_name):
        """Read acquired speed data from the raw data file.

        The columns in the file are:
         - Date as MM/DD/YYYY
         - Time at HH:MM:SS.UUUUUU
         - Microseconds since start of trial
         - Speed in rpm (1 rpm = 50 cm/s), always negative!
         - Unsure; seems to be unused

        The date & time columns give the global experiment time. The third column is used
        to identify where trials begin & end. It will reset both at the end of one trial,
        then again at the start of the next, giving a short period of 'junk' data inbetween.

        :param file_name: path to the file
        :returns: (speed_data, initial_time), where speed_data is the file contents as a
        Pandas data table, and initial_time is the experiment start time, which sets the
        session_start_time for the NWB file.
        """
        self.log('Loading speed data from {}', file_name)
        assert os.path.isfile(file_name)
        speed_data = pd.read_csv(file_name, sep='\t', header=None, usecols=[0, 1, 2, 3], index_col=0,
                                 names=('Date', 'Time', 'Trial time', 'Speed'),
                                 dtype={'Trial time': np.int32, 'Speed': np.float32},
                                 parse_dates=[[0, 1]],  # Combine first two cols
                                 dayfirst=True, infer_datetime_format=True,
                                 memory_map=True)
        initial_offset = pd.Timedelta(microseconds=speed_data['Trial time'][0])
        initial_time = speed_data.index[0] - initial_offset
        return speed_data, initial_time

    def add_speed_data(self, speed_data, initial_time):
        """Add acquired speed data the the NWB file.

        Creates the /acquisition/speed_data and
        /acquisition/trial_times groups.

        :param speed_data: raw speed data loaded from file by read_speed_data(), as a
        Pandas data table
        :param initial_time: experiment start time, from read_speed_data()
        """
        rel_times = (speed_data.index - initial_time).total_seconds().values
        ts_attrs = {'description': 'Raw mouse speed data.',
                    'comments': 'Speed is in rpm, with conversion factor to cm/s specified.'}
        speed_attrs = {'unit': 'cm/s', 'conversion': 50.0 / 60.0, 'resolution': 0.001 * 50 / 60}
        time_attrs = {'unit': 'second', 'conversion': 1e6, 'resolution': 1e-6}
        self.add_time_series_data('speed_data', speed_data['Speed'].values, rel_times,
                                  ts_attrs=ts_attrs, data_attrs=speed_attrs)
        ts_attrs['description'] = 'Per-trial times for mouse speed data.'
        self.add_time_series_data('trial_times', speed_data['Trial time'].values, rel_times,
                                  ts_attrs=ts_attrs, data_attrs=time_attrs)
        self._write()

    def get_times(self, timeseries):
        """Get the timestamps for a timeseries as a numpy array.

        Will handle both the case where there is a 'timestamps' attribute, and the case where
        these must be determined from 'starting_time' and rate.

        :param timeseries: the timeseries
        """
        if timeseries.timestamps is not None:
            t = np.array(timeseries.timestamps)
        else:
            n = timeseries.num_samples
            t0 = timeseries.starting_time
            rate = timeseries.rate
            t = t0 + np.arange(n) / rate
        return t

    def determine_trial_times(self):
        """Use the acquired speed data to determine the start & end times for each trial.

        Each trial will be defined as an epoch within the NWB file, and the relevant portions
        of the speed data linked to these.  The epochs will be named 'trial_0001' etc.

        Trials are identified by the resets in the 'trial_times' timeseries.  This stores
        the relative times for each speed reading within a single trial.  Hence when it
        resets (i.e. one entry is less than the one before) this marks the end of a trial.
        There is a short interval between trials which still has speed data recorded, so it's
        the second reset which marks the start of the next trial.
        """
        speed_data_ts = self.nwb_file.get_acquisition('speed_data')
        if self.labview_version is LabViewVersions.pre2018:
            self.log('Calculating trial times from speed data')
            trial_times_ts = self.nwb_file.get_acquisition('trial_times')
            trial_times = np.array(trial_times_ts.data)
            # Prepend -1 so we pick up the first trial start
            # Append -1 in case there isn't a reset recorded at the end of the last trial
            deltas = np.ediff1d(trial_times, to_begin=-1, to_end=-1)
            # Find resets and pair these up to mark start & end points
            reset_idxs = (deltas < 0).nonzero()[0].copy()
            assert reset_idxs.ndim == 1
            num_trials = reset_idxs.size // 2  # Drop the extra reset added at the end if
            reset_idxs = np.resize(reset_idxs, (num_trials, 2))  # it's not needed
            reset_idxs[:, 1] -= 1  # Select end of previous segment, not start of next
            # Index the timestamps to find the actual start & end times of each trial. The start
            # time is calculated using the offset value in the first reading within the trial.
            rel_times = self.get_times(trial_times_ts)
            epoch_times = rel_times[reset_idxs]
            epoch_times[:, 0] -= trial_times[reset_idxs[:, 0]] * 1e-6
        elif self.labview_version is LabViewVersions.v231:
            epoch_times = self.trial_times
        # Create the epochs in the NWB file
        # Note that we cannot pass the actual start time to nwb_file.add_epoch since it
        # would add the last previous junk speed reading to the start of the next trial,
        # since they have exactly the same timestamp. We therefore cheat and pass a time
        # point 1 ns after that time, instead. All equipment records times with >1us
        # resolution, so this shouldn't be a problem. The variable start_time is therefore
        # maybe better thought of as the time of the last junk speed reading.
        # We also massage the end time since otherwise data points at exactly that time are
        # omitted.
        self.nwb_file.add_epoch_column('epoch_name', 'the name of the epoch')
        for i, (start_time, stop_time) in enumerate(epoch_times):
            assert stop_time > start_time >= 0
            trial = 'trial_{:04d}'.format(i + 1)
            self.nwb_file.add_epoch(
                epoch_name=trial,
                start_time=start_time if i == 0 else start_time + 1e-9,
                stop_time=stop_time + 1e-9,
                timeseries=[speed_data_ts])
            # We also record exact start & end times in the trial table, since our epochs
            # correspond to trials.
            self.nwb_file.add_trial(start_time=start_time, stop_time=stop_time)
        self._write()

    def add_stimulus(self):
        """Add information about the stimulus presented.

        This is taken from the metadata.yaml at present, since Labview does not record this
        information.

        This adds a TimeSeries group to /stimulus/presentation containing the timings of air
        puffs. At present the stimulus is always presented at the same time within each trial,
        so the only argument is the delay from the start of the trial until the puff. The 'data'
        for this time series is simply the text 'puff' at each occasion.
        """
        for stim in self.experiment['stimulus_details']:
            attrs = {
                'name': stim['name'],
                'description': stim['description'],
                'comments': stim['comments'],
                'unit': 'n/a',
                'conversion': float('nan'),
                'resolution': float('nan')
            }
            # TODO We can maybe build puffs and times a bit more efficiently or
            # in fewer steps, although it probably won't make a huge difference
            # (eg we can now get all the times with a single indexing expression)
            num_epochs = len(self.nwb_file.epochs)
            puffs = [u'puff'] * num_epochs
            times = np.zeros((len(puffs),), dtype=np.float64)
            for i in range(num_epochs):
                times[i] = self.nwb_file.epochs[i, 'start_time'] + stim['trial_time_offset']
            attrs['timestamps'] = times
            attrs['data'] = puffs
            self.nwb_file.add_stimulus(TimeSeries(**attrs))
        self._write()

    def read_cycle_relative_times(self, folder_path):
        """Read the files containing relative times and store the values in memory.

        Depending on the version of software used, the file is named either 'Single cycle relative times.txt' or
        'Single cycle relative times_HW.txt'. This also requires reading the ROI.dat file.

        Note that while the file has times in microseconds, we convert to seconds for consistency
        with other timestamps in NWB.
        """

        def rel(file_name):
            """Return the path of a file name relative to the Labview folder."""
            return os.path.join(folder_path, file_name)

        roi_path = rel('ROI.dat')
        assert os.path.isfile(roi_path)
        if self.labview_version is LabViewVersions.pre2018:
            file_path = rel('Single cycle relative times.txt')
            assert os.path.isfile(file_path)
            timings = LabViewTimingsPre2018(relative_times_path=file_path,
                                            roi_path=roi_path,
                                            dwell_time=self.imaging_info.dwell_time / 1e6)
        elif self.labview_version is LabViewVersions.v231:
            file_path = rel('Single cycle relative times_HW.txt')
            assert os.path.isfile(file_path)
            timings = LabViewTimings231(relative_times_path=file_path,
                                        roi_path=roi_path,
                                        n_cycles_per_trial=self.imaging_info.cycles_per_trial,
                                        n_trials=len(self.trial_times),
                                        dwell_time=self.imaging_info.dwell_time / 1e6)
        else:
            raise ValueError('Unsupported LabView version for timings {}.'.format(self.labview_version))
        self.raw_pixel_time_offsets = timings.pixel_time_offsets
        self.cycle_time = timings.cycle_time

    def read_functional_data(self, folder_path):
        """Import functional data from Labview TDMS files.

        The folder contains files named like like "NNN.tdms", where NNN gives the trial number.
        These files are thus per-trial, but a TimeSeries needs to contain all trials. We link to
        portions corresponding to single trials from epochs.

        Each file contains data for 2 channels in the group 'Functional Imaging Data', where
        'Channel 0 Data' is Red and 'Channel 1 Data' is Green. Each channel contains data for all
        pixels in all ROIs for all time within that trial, as a single 1d array. Within this array,
        we have data first for all pixels in the first ROI at time 0, then the second ROI at time
        0, and so on through all ROIs, before moving to data from the next cycle. For 2d ROIs, it
        scans first over the X dimension then over Y.

        While it might seem that this data is well suited to become RoiResponseSeries within
        /processing/Acquired_ROIs/Fluoresence, that time series type assumes a single value per
        ROI per time, which doesn't support storing raw data from multi-pixel ROIs. Instead,
        the data will be stored within /acquisition, in TwoPhotonSeries named like ROI_NNN_Green,
        where NNN is the global (not per-imaging-plane) ROI number.

        We define the ROIs within their imaging planes first, storing zeros for the data, then
        read each TDMS file to fill in real recordings. The channel data will have its dimensions
        permuted to match the NWB (t, z, y, x) arrangement, and be stored in the appropriate
        segment of the timeseries. At present all ROIs are 2d (or less, but can be represented as
        such with length 1 dimensions), however this structure allows for extension to 3d in the
        future.

        The single cycle time (self.cycle_time) gives you the difference
        in acquisition time between successive lines in a file. Time starts at the beginning of
        the trial (epoch in NWB speak). This is the time to record in the timestamps field. Note
        that even though these are evenly spaced we can't use starting_time and rate, since this
        would not account for time between trials.
        """
        self.log('Loading functional data from {}', folder_path)
        assert os.path.isdir(folder_path)
        # Figure out timestamps, measured in seconds
        epoch_names = self.nwb_file.epochs[:, 'epoch_name']
        trials = [int(s[6:]) for s in epoch_names]  # names start with 'trial_'
        cycles_per_trial = self.imaging_info.cycles_per_trial
        num_times = cycles_per_trial * len(epoch_names)
        single_trial_times = np.arange(cycles_per_trial) * self.cycle_time
        times = np.zeros((num_times,), dtype=float)
        # TODO Perhaps this loop can be vectorised
        for i in range(len(epoch_names)):
            trial_start = self.nwb_file.epochs[i, 'start_time']
            times[i * cycles_per_trial:
                  (i + 1) * cycles_per_trial] = single_trial_times + trial_start
        self.custom_silverlab_dict['cycle_time'] = self.cycle_time
        self.custom_silverlab_dict['cycles_per_trial'] = cycles_per_trial

        # We now know all we need to write the custom part of Silver Lab data
        self.add_custom_silverlab_data()

        # Prepare attributes for timeseries groups and datasets (common to all instances)
        data_attrs = {'unit': 'intensity', 'conversion': 1.0, 'resolution': float('NaN')}
        ts_desc_template = 'Fluorescence data acquired from the {channel} channel in {roi_name}.'
        ts_attrs = {'comments': 'The AOL microscope can acquire just the pixels comprising defined'
                                ' ROIs. This timeseries records those pixels over time for a'
                                ' single ROI & channel.'}
        PixelTimeOffsets = get_class('PixelTimeOffsets', 'silverlab_extended_schema')
        ROISeriesWithPixelTimeOffsets = get_class('ROISeriesWithPixelTimeOffsets', 'silverlab_extended_schema')

        gains = self.imaging_info.gains
        # Iterate over ROIs, which are nested inside each imaging plane section
        all_rois = {}
        seg_iface = self.nwb_file.processing['Acquired_ROIs'].get("ImageSegmentation")
        for plane_name, plane in seg_iface.plane_segmentations.items():
            self.log('  Defining ROIs for plane {}', plane_name)
            # ROIs are added using an integer id, but they are retrieved using
            # their index into the table, which does not necessarily correspond
            # to the id (in our case, ids start from 1 and indices from 0). See:
            # https://github.com/NeurodataWithoutBorders/pynwb/issues/673
            for roi_num, roi_ind in self.roi_mapping[plane_name].items():
                roi_name = 'ROI_{:03d}'.format(roi_num)
                if roi_num not in all_rois.keys():
                    all_rois[roi_num] = {}
                for ch, channel in {'A': 'Red', 'B': 'Green'}.items():
                    # Set zero data for now; we'll read the real data later
                    # TODO: The TDMS uses 64 bit floats; we may not really need that precision!
                    # The exported data seems to be rounded to unsigned ints. Issue #15.
                    roi_dimensions = plane[roi_ind, 'dimensions']
                    data_shape = np.concatenate((roi_dimensions, [num_times]))[::-1]
                    data = np.zeros(data_shape, dtype=np.float64)
                    # Create the timeseries object and fill in standard metadata
                    ts_name = 'ROI_{:03d}_{}'.format(roi_num, channel)
                    ts_attrs['description'] = ts_desc_template.format(channel=channel.lower(),
                                                                      roi_name=roi_name)
                    data_attrs['dimension'] = roi_dimensions
                    data_attrs['format'] = 'raw'
                    pixel_size_in_m = self.imaging_info.field_of_view / 1e6 / self.imaging_info.frame_size
                    data_attrs['field_of_view'] = roi_dimensions * pixel_size_in_m
                    data_attrs['imaging_plane'] = plane.imaging_plane
                    data_attrs['pmt_gain'] = gains[channel]
                    data_attrs['scan_line_rate'] = 1 / self.cycle_time
                    # TODO The below are not supported, so will require an extension
                    # However, they can be extracted by the name of the TimeSeries
                    # or by looking into the corresponding ROI.
                    # ts.set_custom_dataset('roi_name', roi_name)
                    # ts.set_custom_dataset('channel', channel)
                    # # Save the time offset(s) for this ROI, as a link
                    # ts.set_dataset('pixel_time_offsets', 'link:' + roi['pixel_time_offsets'].name)
                    # We need to retrieve the overall order of the ROI (not just
                    # within this plane). This is given by roi_num but starts at 1.
                    data_attrs['pixel_time_offsets'] = PixelTimeOffsets(self.raw_pixel_time_offsets[roi_num - 1])
                    self.add_time_series_data(ts_name, data=data, times=times,
                                              kind=ROISeriesWithPixelTimeOffsets,
                                              ts_attrs=ts_attrs, data_attrs=data_attrs)

                    # Store the path where these data should go in the file
                    all_rois[roi_num][channel] = '/acquisition/{}/data'.format(ts_name)
                    # Link to these data within the epochs
                    # TODO This will need converting to the new API, which is less flexible.
                    # Do we need to add these timeseries before we create the epochs?
                    # This is kind of a chicken-and-egg problem, as we need the
                    # epoch information for the above calculations (although it
                    # does need not to be stored in the NWBFile epochs yet)
                    # for trial, epoch_name in enumerate(epoch_names):
                    #     epoch = self.nwb_file.get_node('/epochs/' + epoch_name)
                    #     series_ref_in_epoch = epoch.make_group('<timeseries_X>', ts_name)
                    #     series_ref_in_epoch.set_dataset('idx_start', trial * cycles_per_trial)
                    #     series_ref_in_epoch.set_dataset('count', cycles_per_trial)
                    #     series_ref_in_epoch.make_group('timeseries', ts)
        # We need to write the zero-valued timeseries before editing them!
        self._write()
        # The shape to put the TDMS data in for more convenient indexing
        # TODO Are the roi_dimensions always the same across ROIs? (it seems that
        # this was the implication from the previous version of the code, as it
        # always used the last value of roi_dimensions - but that may be a bug?)
        ch_data_shape = np.concatenate((roi_dimensions,
                                        [len(all_rois), cycles_per_trial]))[::-1]
        self._write_roi_data(all_rois, len(trials), cycles_per_trial, ch_data_shape, folder_path)

    def add_custom_silverlab_data(self, include_opto=True):
        metadata_class = get_class('SilverLabMetaData', 'silverlab_extended_schema')
        custom_metadata = {
            'name': 'silverlab_metadata',
            'silverlab_api_version': self.SILVERLAB_NWB_VERSION
        }
        # If creating from metadata only, we don't have any LabVIEW data
        if self.labview_version:
            custom_metadata['labview_version'] = self.labview_version.value
        silverlab_metadata = metadata_class(**custom_metadata)
        self.nwb_file.add_lab_meta_data(silverlab_metadata)
        if include_opto:
            optophysiology_class = get_class('SilverLabOptophysiology', 'silverlab_extended_schema')
            silverlab_optophysiology = optophysiology_class(
                name='silverlab_optophysiology',
                cycle_time=self.custom_silverlab_dict['cycle_time'],
                cycles_per_trial=self.custom_silverlab_dict[
                    'cycles_per_trial'],
                frame_size=self.custom_silverlab_dict['frame_size'],
                imaging_mode=self.custom_silverlab_dict['imaging_mode'],
                pockels=self.custom_silverlab_dict['zplane_pockels']
            )
            self.nwb_file.add_lab_meta_data(silverlab_optophysiology)
        self._write()

    def _write_roi_data(self, all_rois, num_trials, cycles_per_trial,
                        ch_data_shape, folder_path):
        """Edit the NWB file directly to add the real ROI data."""
        with h5py.File(self.nwb_path, 'a') as out_file:
            # Iterate over trials, reading data from the TDMS file for each
            for trial_index in range(num_trials):
                self.log('  Reading TDMS {}', trial_index + 1)
                file_path = os.path.join(folder_path, '{:03d}.tdms'.format(trial_index + 1))
                tdms_file = TdmsFile(file_path,
                                     memmap_dir=tempfile.gettempdir())
                time_segment = slice(trial_index * cycles_per_trial,
                                     (trial_index + 1) * cycles_per_trial)
                for ch, channel in {'0': 'Red', '1': 'Green'}.items():
                    # Reshape the TDMS data into an nd array
                    # TODO: Consider precision: the round() here is to match the exported data...
                    ch_data = np.round(tdms_file.channel_data('Functional Imaging Data',
                                                              'Channel {} Data'.format(ch)))
                    ch_data = ch_data.reshape(ch_data_shape)
                    # Copy each ROI's data into the NWB
                    for roi_num, data_paths in all_rois.items():
                        channel_path = out_file[data_paths[channel]]
                        channel_path[time_segment, ...] = ch_data[:, roi_num - 1, ...]
        # Update our reference to the NWB file, since it's now out of sync
        # We need to keep a reference to the IO object, as the file contents are
        # not read until needed
        self.nwb_io = NWBHDF5IO(self.nwb_path, 'r')
        self.nwb_file = self.nwb_io.read()
        # Quick fix...
        self.nwb_io.close()

    def add_imaging_plane(self, name, description, origin_coords, grid_spacing,
                          green=True, red=True):
        """Add a new imaging plane definition to /general/optophysiology.

        :param name: A name for the NWB group representing this imaging plane.
        :param description: Brief text description of the plane, e.g. "Reference Z stack",
            "Pointing mode acquisition sequence".
        :param origin_coords: physical location of the first element of the imaging plane
        :param grid_spacing: distance along each axis between two adjacent pixels in the plane (assumes a regular grid)
        :param green: Whether to include the green channel.
        :param red: Whether to include the red channel.
        """
        opto_metadata = self.experiment['optophysiology']
        cycle_rate = 1 / self.cycle_time  # Hz
        channels = []
        if green:
            channel = OpticalChannel('green',
                                     description='Green channel, typically used for active signal.',
                                     emission_lambda=float(opto_metadata['emission_lambda']['green']))
            channels.append(channel)
        if red:
            channel = OpticalChannel('red',
                                     description='Red channel, typically used for reference.',
                                     emission_lambda=float(opto_metadata['emission_lambda']['red']))
            channels.append(channel)
        self.nwb_file.create_imaging_plane(
            name=name,
            optical_channel=channels,
            description=description,
            device=self.nwb_file.devices['AOL_microscope'],
            excitation_lambda=float(opto_metadata['excitation_lambda']),
            imaging_rate=cycle_rate,
            indicator=opto_metadata['calcium_indicator'],
            location=opto_metadata['location'],
            origin_coords=origin_coords,
            origin_coords_unit='micrometers',
            grid_spacing=grid_spacing,
            grid_spacing_unit='micrometers',
            reference_frame='TODO: In lab book (partly?)'
        )

    def read_zplane(self, zplane_path):
        """Determine coordinates of reference image stack from Zplane_Pockels_Values.dat.

        This also uses information from the LabView .ini file to define image planes etc
        in /general/optophysiology.

        The .dat file has 4 columns: Z offset from focal plane (micrometres), normalised Z,
        'Pockels' i.e. laser power in %, and z offset for drive motors. We save this raw
        array as the extension dataset /general/optophysiology/zplane_pockels.

        Also sets up self.zplanes as a map from Z coordinate (in microns) to imaging plane name.
        """
        self.log('Loading imaging plane information from {}', zplane_path)
        assert os.path.isfile(zplane_path)
        zplane_data = pd.read_csv(
            zplane_path, sep='\t', skiprows=2, skip_blank_lines=True,
            names=('z', 'z_norm', 'laser_power', 'z_motor'), header=0,
            index_col=False)
        num_pixels = self.imaging_info.frame_size
        plane_width_in_microns = self.imaging_info.field_of_view
        self.zplanes = {}
        for plane in zplane_data.itertuples():
            name = 'Zstack{:04d}'.format(plane.Index + 1)
            self.zplanes[plane.z] = name
            # TODO: Is it OK to have 2D spacing with 3D coords (allowed by pynwb, but this doesn't mean it's OK)?
            spacing = plane_width_in_microns / (num_pixels - 1)
            origin_coords = [0, 0, plane.z]
            self.add_imaging_plane(
                name=name,
                description='Reference Z stack',
                origin_coords=origin_coords,
                grid_spacing=np.float32([spacing, spacing, 0])
            )
        ZplanePockelsDatasetClass = get_class('ZplanePockelsDataset', 'silverlab_extended_schema')
        self.custom_silverlab_dict['zplane_pockels'] = ZplanePockelsDatasetClass(
            columns=zplane_data.columns.tolist(),
            data=zplane_data.values)
        self.custom_silverlab_dict['frame_size'] = [num_pixels, num_pixels]
        self._write()

    def read_zstack(self, zstack_folder):
        """Add the reference Z stack images into /acquisition.

        The folder holds one .tif file per imaging plane per channel, with the planes ordered
        as in the Zplane_Pockels_Values.dat file (see read_zplane) and hence the order matches
        the ZstackNNNN planes added there. The file names are like GreenChannel_0001.tif.

        We create a single-image TwoPhotonSeries for each plane for each channel in
        /acquisition/Zstack_<channel>_<plane>.

        Also fills in self.zstack as a mapping from [plane_name][channel_name] to the
        corresponding acquisition name (Zstack_<channel_name>_<plane_name>).
        """
        self.log('Loading reference Z stack from {}', zstack_folder)
        assert os.path.isdir(zstack_folder)
        cycle_rate = 1 / self.cycle_time  # Hz
        self.zstack = {}
        for plane_name, plane in self.nwb_file.imaging_planes.items():
            assert plane_name.startswith('Zstack'), 'Found unexpected plane {}'.format(plane_name)
            self.zstack[plane_name] = {}
            for channel in ('Green', 'Red'):
                plane_index = plane_name[6:]
                group_name = 'Zstack_{}_{}'.format(channel, plane_index)
                file_path = os.path.join(zstack_folder,
                                         channel + 'Channel_' + plane_index + '.tif')
                if not os.path.isfile(file_path):
                    print('Expected Zstack file "{}" missing; skipping.'.format(file_path))
                    continue
                img = tifffile.imread(file_path)
                num_pixels = self.imaging_info.frame_size
                width_in_metres = self.imaging_info.field_of_view / 1e6
                # Save img to NWB
                ts_attrs = {'description': 'Initial reference Z stack plane',
                            'comments': 'Contains single slice from {} channel'.format(
                                channel.lower())}
                data_attrs = {'unit': 'intensity', 'conversion': 1.0,
                              'resolution': float('NaN'),
                              'dimension': [num_pixels, num_pixels],
                              'format': 'tiff',
                              'field_of_view': [width_in_metres, width_in_metres],
                              'imaging_plane': plane,
                              'pmt_gain': self.imaging_info.gains[channel],
                              'scan_line_rate': cycle_rate,
                              # TODO A TwoPhotonSeries doesn't store channel information.
                              # We can either an extension of it, but the channel is also
                              # stored in the imaging plane (linked to from the Series),
                              # so perhaps we don't need to?
                              # ts.set_custom_dataset('channel', channel)
                              }
                self.add_time_series_data(group_name, data=[img], times=np.array([0.0]),
                                          kind=TwoPhotonSeries,
                                          ts_attrs=ts_attrs, data_attrs=data_attrs)
                # TODO Since this is only used when adding ROIs, it might be better
                # to have a method that returns the acquisition name, rather than
                # store the mapping.
                self.zstack[plane_name][channel] = group_name
        self._write()

    def add_rois(self, roi_path):
        """Add the locations of ROIs as an ImageSegmentation module.

        We read a ROI.dat file to determine ROI locations. This has many tab-separated columns:
            ROI index; ROI ID; ROI Time (ns); Pixels in ROI;
            X start; Y start; Z start; X stop; Y stop; Z stop;
            Angle (deg); Composite ID; Number of lines; Frame Size; Zoom;
            Laser Power (%); ROI group ID.

        Each ROI must lie within one of the Z planes in the reference Z stack. We therefore can
        represent it as a rectangle (size 1x1 for pointing mode) within that plane. The relevant
        slice slice from the Z stack is used as the reference image. We group the ROIs by Z
        coordinate, since each imaging plane should only be listed once in the ImageSegmentation
        module; within a given plane, the original relative ordering of ROIs is maintained.

        TODO: Consider adding an array dataset of object references to the ROI definitions - see
        http://docs.h5py.org/en/latest/refs.html for details of how to do this. Would enable quick
        access to all ROIs (in the defined order) without having to iterate over imaging planes
        then sort. It's less of an issue with the timeseries ROI data, since that's in groups
        organised by ROI number and channel name, so we can iterate there. Issue #16.
        """
        self.log('Loading ROI locations from {}', roi_path)
        assert os.path.isfile(roi_path)
        roi_data = pd.read_csv(
            roi_path, sep='\t', header=0, index_col=False, dtype=np.float16,
            converters={'Z start': np.float64, 'Z stop': np.float64}, memory_map=True)
        # Rename the columns so that we can use them as identifiers later on
        column_mapping = {
            'ROI index': 'roi_index', 'Pixels in ROI': 'num_pixels',
            'X start': 'x_start', 'Y start': 'y_start', 'Z start': 'z_start',
            'X stop': 'x_stop', 'Y stop': 'y_stop', 'Z stop': 'z_stop',
            'Laser Power (%)': 'laser_power', 'ROI Time (ns)': 'roi_time_ns',
            'Angle (deg)': 'angle_deg', 'Composite ID': 'composite_id',
            'Number of lines': 'num_lines', 'Frame Size': 'frame_size',
            'Zoom': 'zoom', 'ROI group ID': 'roi_group_id'
        }
        roi_data.rename(columns=column_mapping, inplace=True)
        module = self.nwb_file.create_processing_module(
            'Acquired_ROIs',
            'ROI locations and acquired fluorescence readings made directly by the AOL microscope.')
        # Convert some columns to int
        roi_data = roi_data.astype(
            {'x_start': np.uint16, 'x_stop': np.uint16, 'y_start': np.uint16, 'y_stop': np.uint16,
             'num_pixels': int})
        seg_iface = ImageSegmentation()
        module.add(seg_iface)
        self._write()
        # Define the properties of the imaging plane itself, if not a Z plane
        self.custom_silverlab_dict['imaging_mode'] = self.mode.name
        if self.mode is Modes.pointing:
            # Sanity check that each ROI is a single pixel
            assert np.all(roi_data.num_pixels == 1)
        # Figure out which plane each ROI is in
        assert (roi_data['z_start'] == roi_data['z_stop']).all()  # Planes are flat in Z
        grouped = roi_data.groupby('z_start', sort=False)
        # Iterate over planes and define ROIs
        self.roi_mapping = {}  # mapping from ROI ID to row index (used to look up ROIs)
        for plane_z, roi_group in grouped:
            plane_name = self.zplanes[plane_z]
            plane_obj = self.nwb_file.imaging_planes[plane_name]
            plane = seg_iface.create_plane_segmentation(
                description=plane_obj.description,
                imaging_plane=plane_obj,
                name=plane_name,
                reference_images=self.nwb_file.acquisition[self.zstack[plane_name]['Red']]
            )
            # Specify the non-standard data we will be storing for each ROI,
            # which includes all the raw data fields from the original file
            plane.add_column('dimensions', 'Dimensions of the ROI')
            for old_name, new_name in column_mapping.items():
                plane.add_column(new_name, old_name)
            index = 0  # index of the row as it will be stored in the ROI table
            self.roi_mapping[plane_name] = {}
            for row in roi_group.itertuples():
                roi_id = int(row.roi_index)
                # The ROI mask only gives x & y coordinates - z is defined by the imaging plane.
                # The coordinates are also relative to the imaging plane, not absolute. However, our
                # plane coordinates run from 0 to frame_size, so that's easy to compute.
                # The third dimension in the pixels array indicates weight.
                pixels = np.zeros((row.num_pixels, 3), dtype=np.uint16)
                # Pixels are located contiguously from start to stop coordinates.
                num_x_pixels = row.x_stop - row.x_start
                num_y_pixels = row.y_stop - row.y_start
                if self.mode is Modes.pointing:
                    assert row.num_pixels == 1, 'Unexpectedly large ROI in pointing mode'
                    num_x_pixels = num_y_pixels = 1
                assert row.num_pixels == num_x_pixels * num_y_pixels, (
                    'ROI is not rectangular: {} != {} * {}'.format(
                        row.num_pixels, num_x_pixels, num_y_pixels))
                # Record the ROI dimensions for ease of lookup when adding functional data
                dimensions = np.array([num_x_pixels, num_y_pixels], dtype=np.int32)
                for i in range(row.num_pixels):
                    pixels[i, 0] = row.x_start + (i % num_x_pixels)
                    pixels[i, 1] = row.y_start + (i // num_x_pixels)
                    pixels[i, 2] = 1  # weight for this pixel
                plane.add_roi(id=roi_id, pixel_mask=[tuple(r) for r in pixels.tolist()],
                              dimensions=dimensions,
                              **{field: getattr(row, field) for field in column_mapping.values()})
                self.roi_mapping[plane_name][roi_id] = index
                index += 1
        self._write()

    def read_video_data(self, folder_path):
        """Link to video data stored in the given folder.

        :param folder_path: folder containing videos of the experiment.

        Multiple video angles are supported. This method looks for files named like
        '<Base>Cam-relative times.txt' in the folder, each of which contains timing data for a
        single video timeseries. These files contain two tab-separated columns, the first being
        frame numbers, and the second time offsets from the start of the experiment in
        milliseconds.

        The video data itself is stored in .avi files, potentially more than one per camera,
        named like '<Base>Cam-<N>.avi', where the file number <N> starts at 1.

        This method adds an ImageSeries in /acquisition for each camera, linking to
        the existing .avi files with relative paths. The timeseries are named '<Base>Cam'.

        Since we only add a link, we do not compress the data here when writing.
        """
        if av is None:
            raise ValueError('Unable to read video data without the av library installed')
        # Quick fix?
        with NWBHDF5IO(self.nwb_path, 'a') as io:
            self.nwb_file = io.read()
            self.log('Loading video data from {}', folder_path)
            assert os.path.isdir(folder_path)
            nwb_dir = os.path.dirname(os.path.realpath(self.nwb_path))
            timing_suffix = '-relative times.txt'
            timing_files = glob.glob(os.path.join(folder_path, '*' + timing_suffix))
            for timing_file_path in timing_files:
                cam_name = os.path.basename(timing_file_path)[:-len(timing_suffix)]
                self.log('Camera: {}', cam_name)
                frame_rel_times = pd.read_csv(timing_file_path, sep='\t', names=('Frame', 'RelTime'))
                frame_rel_times['RelTime'] *= 1e-3  # Convert to seconds
                # Determine properties of each .avi file
                avi_files = glob.glob(os.path.join(folder_path, cam_name + '-*.avi'))
                num_frames = np.zeros((len(avi_files),), dtype=np.int64)
                video_file_paths = [''] * len(avi_files)
                for avi_file in avi_files:
                    file_name = os.path.basename(avi_file)
                    self.log('Video: {}', file_name)
                    index = int(file_name[len(cam_name) + 1:-4]) - 1
                    avi_file = os.path.realpath(avi_file)
                    try:
                        video_file_paths[index] = os.path.relpath(avi_file, nwb_dir)
                    except ValueError:
                        # Particularly on Windows, it's sometimes impossible to construct
                        # a relative path, so fall back to absolute
                        video_file_paths[index] = avi_file
                    container = av.open(avi_file)
                    vid = container.streams.video[0]
                    num_frames[index] = vid.frames
                    if index == 0:
                        vid_rate = vid.rate
                        vid_dimensions = [vid.width, vid.height]
                    del container, vid
                starting_frames = np.roll(np.cumsum(num_frames), 1)
                starting_frames[0] = 0
                # Add camera to list of devices
                self.nwb_file.create_device(
                    cam_name, self.user_metadata['devices'][cam_name])
                # Create timeseries
                ts_attrs = {
                    'description': 'Video recording of mouse behaviour.',
                    'comments': 'Frame rate {} fps'.format(vid_rate)
                }
                data_attrs = {
                    'format': 'external',
                    'external_file': video_file_paths,
                    'starting_frame': starting_frames,
                    'dimension': vid_dimensions,
                }
                self.add_time_series_data(
                    cam_name, data=None, times=frame_rel_times['RelTime'].values,
                    ts_attrs=ts_attrs, data_attrs=data_attrs, kind=ImageSeries)
            io.write(self.nwb_file)

    def _write(self):
        with NWBHDF5IO(self.nwb_path, 'w') as io:
            io.write(self.nwb_file)



'''
Command-line entrypoints for the Silver Lab pipeline.
'''

import argparse

from .metadata_gui import run_editor
from .nwb_file import NwbFile
from .signature import SignatureGenerator


def import_labview():
    """Command line script to import a Labview folder to NWB format."""
    parser = argparse.ArgumentParser(description='Import Labview to NWB.')
    parser.add_argument('nwb_path',
                        help='path to the NWB file to write')
    parser.add_argument('labview_path',
                        help='path to the Labview folder to import')
    parser.add_argument('--no-gui', dest='gui', action='store_false',
                        help="don't show a GUI to edit metadata; just reuse the last session")
    parser.add_argument('--check-sig', action='store_true',
                        help="check the generated NWB file against an existing signature."
                        " This expects a .sig file to be present next to the .nwb file.")
    args = parser.parse_args()

    if args.gui:
        run_editor()

    with NwbFile(args.nwb_path, mode='w') as nwb:
        nwb.import_labview_folder(args.labview_path)

    if args.check_sig:
        import os
        sig_path = os.path.splitext(args.nwb_path)[0] + '.sig'
        if os.path.exists(sig_path):
            sig_gen = SignatureGenerator()
            print('Comparing against signature file {}'.format(sig_path))
            sig_gen.compare_to_sig(args.nwb_path, sig_path)
        else:
            parser.error('No signature file {} found'.format(sig_path))


def import_metadata():
    """Command line script to create a bare-bones NWB file from metadata."""
    parser = argparse.ArgumentParser(description='Import metadata to NWB.')
    parser.add_argument('nwb_path',
                        help='path to the NWB file to write')
    parser.add_argument('metadata_path',
                        help='path to a YAML metadata file')
    parser.add_argument('user',
                        help='name of a user whose experiment to import (optional)',
                        nargs='?',
                        default=None)
    args = parser.parse_args()

    with NwbFile(args.nwb_path, mode='w') as nwb:
        nwb.create_from_metadata(args.metadata_path, args.user)



from pynwb.spec import NWBAttributeSpec, NWBDatasetSpec, NWBGroupSpec, NWBNamespaceBuilder


def generate_extended_schema():
    # set up silverlab namespace
    ns_builder = NWBNamespaceBuilder('Extensions for acousto-optic lens data',
                                     'silverlab_extended_schema',
                                     'Silver lab data extension to NWB format for acousto-optic lens experiments',
                                     version='0.4')
    ns_builder.include_type('LabMetaData', namespace='core')
    ns_builder.include_type('TwoPhotonSeries', namespace='core')

    # define attributes Silver lab extension
    cycle_time_attr = NWBAttributeSpec(name='cycle_time',
                                       doc='time in seconds for the microscope to acquire all ROIs once '
                                       'and return to its starting position',
                                       dtype='float')
    cycles_per_trial_attr = NWBAttributeSpec(name='cycles_per_trial',
                                             doc='how many microscope cycles occur in each experimental trial',
                                             dtype="int")
    imaging_mode_attr = NWBAttributeSpec(name='imaging_mode',
                                         doc='the acquisition mode for the experiment; '
                                         'pointing = single-voxel ROIs, '
                                         'miniscan = 2d rectangular ROIs, '
                                         'volume = 3d cuboid ROIs',
                                         dtype='text')
    frame_size_attr = NWBAttributeSpec(name='frame_size',
                                       doc='the 2d imaging frame size in voxels',
                                       shape=(2,),
                                       dtype='int')
    silverlab_api_version_attr = NWBAttributeSpec(name='silverlab_api_version',
                                                  doc='For potential future backwards compatibility, '
                                                  'store the \'version\' of this API that created the file.',
                                                  dtype='text')
    labview_version_attr = NWBAttributeSpec(name='labview_version',
                                            doc='The version of LabVIEW the data came from',
                                            dtype='text',
                                            required=False)
    pockels_column_names_attr = NWBAttributeSpec(name='columns',
                                                 doc='column names for the zplane pockels dataset',
                                                 shape=(4,),
                                                 dtype='text')
    # define datasets for Silver lab extensions
    zplane_pockels_ds = NWBDatasetSpec(doc='pockels data set, recording calibration data '
                                           'for focusing at different z-planes in four columns: '
                                           'Z offset from focal plane (micrometres), '
                                           'normalised Z, '
                                           '\'Pockels\' i.e. laser power in %, '
                                           'and z offset for drive motors',
                                       name='pockels',
                                       shape=(None, 4),
                                       attributes=[pockels_column_names_attr],
                                       neurodata_type_def='ZplanePockelsDataset')
    # define groups for Silver lab extensions
    silverlab_optophys_specs = NWBGroupSpec(doc='A place to store Silver lab specific optophysiology data',
                                            attributes=[cycle_time_attr,
                                                        cycles_per_trial_attr,
                                                        frame_size_attr,
                                                        imaging_mode_attr],
                                            datasets=[zplane_pockels_ds],
                                            neurodata_type_def='SilverLabOptophysiology',
                                            neurodata_type_inc='LabMetaData')
    silverlab_metadata_specs = NWBGroupSpec(doc='A place to store Silver lab specific metadata',
                                            attributes=[silverlab_api_version_attr, labview_version_attr],
                                            neurodata_type_def='SilverLabMetaData',
                                            neurodata_type_inc='LabMetaData',
                                            )

    # dimensions ordered as t, x, y [, z], like the TimeSeries data itself
    silverlab_pixel_time_offset_data = NWBDatasetSpec(doc='A datastructure to hold time offsets for pixels. The'
                                                          'time offsets are the acquisition time of each pixel '
                                                          'relative to a starting time. The starting time is the '
                                                          'start of the cycle for pre-2018 LabView versions, '
                                                          'and the start of the trial for new versions.',
                                                      name='pixel_time_offsets',
                                                      shape=[(None, None), (None, None, None), (None, None, None, None)],
                                                      neurodata_type_def='PixelTimeOffsets')
    silverlab_roi_image_specs = NWBGroupSpec(doc='An extension to PyNWB\'s TwoPhotonSeries class, designed to hold '
                                                 'pixels from an ROI as well as the PixelTimeOffsets for them.',
                                             datasets=[silverlab_pixel_time_offset_data],
                                             neurodata_type_def='ROISeriesWithPixelTimeOffsets',
                                             neurodata_type_inc='TwoPhotonSeries')

    # export as schema extension
    ext_source = 'silverlab.ophys.yaml'
    ns_builder.add_spec(ext_source, silverlab_optophys_specs)
    ext_source = 'silverlab.metadata.yaml'
    ns_builder.add_spec(ext_source, silverlab_metadata_specs)
    ext_source = 'silverlab.roi.yaml'
    ns_builder.add_spec(ext_source, silverlab_roi_image_specs)
    ns_builder.export('silverlab.namespace.yaml')


if __name__ == '__main__':
    generate_extended_schema()



"""General test fixtures."""

import os

import pytest


@pytest.fixture(scope="session")
def ref_data_dir():
    return os.path.join(os.path.dirname(__file__), 'data')


@pytest.fixture(autouse=True)
def metadata_config(ref_data_dir):
    import silverlabnwb
    silverlabnwb.metadata.set_conf_dir(ref_data_dir)




'''
Handles loading the metadata YAML files.
'''

import collections
import os

import appdirs
import pkg_resources
from ruamel.yaml import YAML


def set_conf_dir(path=appdirs.user_config_dir('SilverLabNwb', 'SilverLab')):
    """Set where to store / read user configuration files.

    Defaults to the system-specific default user configuration folder.

    :param path: path to a configuration folder
    """
    global user_conf_dir, user_conf_path
    user_conf_dir = path
    user_conf_path = os.path.join(user_conf_dir, 'metadata.yaml')


set_conf_dir()  # Set the default paths initially


def read_user_config():
    """Read the user configuration YAML files.

    We first read the default configuration supplied with this package.
    Then we look in the user_config_dir (as defined by appdirs) for any
    machine- & user-specific configuration settings, which override the
    defaults.
    """
    with pkg_resources.resource_stream(__name__, 'metadata.yaml') as defaults:
        settings, comments = read_base_config_file(defaults)
    if os.path.isfile(user_conf_path):
        with open(user_conf_path, 'r') as user_params:
            settings = update_config_file(user_params, settings)
    return settings, comments


def read_custom_config(config_file):
    """Read the metadata from a user-provided YAML file."""
    with open(config_file) as config:
        settings, _ = read_base_config_file(config)
    return settings


def read_base_config_file(stream):
    yaml_reader = YAML()
    parsed_yaml_data = yaml_reader.load(stream)
    yaml_data_comments = read_comments(parsed_yaml_data)
    settings = {}
    recursive_dict_update(settings, strip_strings(parsed_yaml_data))
    return settings, yaml_data_comments


def should_keep_comment(line):
    return len(line) > 0 and not line.startswith('>')


def beautify_comment(comment_token):
    """Format CommentToken instance to display nicely as a tooltip."""
    yaml_comment = comment_token[2].value
    return "\n".join(filter(should_keep_comment, [line.strip() for line in yaml_comment.split('#')]))


def read_comments(settings):
    comments = {}
    for k, v in settings.items():
        if isinstance(v, collections.Mapping):
            comments[k] = read_comments(v)
        elif isinstance(v, list):
            comments[k] = [read_comments(entry) for entry in v]
        else:
            comment = settings.ca.items.get(k, None)
            if comment is not None:
                comments[k] = beautify_comment(comment)
    return comments


def update_config_file(stream, base_settings):
    """Read a single YAML configuration file.

    :param stream: the open file object.
    :param base_settings: a settings dictionary to copy and update
    with the contents of this file.
    """
    yaml_reader = YAML(typ='safe')
    parsed_yaml_data = yaml_reader.load(stream)
    settings = base_settings.copy()
    if parsed_yaml_data is not None:  # if the streamed file is not empty
        recursive_dict_update(settings, strip_strings(parsed_yaml_data))
    return settings


def save_config_file(settings, yaml_instance=None):
    """Save a YAML user configuration file to the default path.

    Will try to create the user's configuration folder if it doesn't exist.

    :param settings: the settings dictionary to save
    :param yaml_instance: a specific YAML instance to use (we may require it to have certain representers)
    """
    if not os.path.isdir(user_conf_dir):
        os.makedirs(user_conf_dir)
    with open(user_conf_path, 'w') as user_params:
        if yaml_instance is None:
            yaml_instance = YAML(typ='safe')
        yaml_instance.dump(settings, user_params)


def recursive_dict_update(base_settings, new_settings):
    """Recursively update settings dictionaries.

    Since our settings are nested dictionaries, we need to merge sub-levels rather
    than overwriting when merging new settings. This method essentially does a
    recursive base_settings.update(new_settings).
    """
    for k, v in new_settings.items():
        if isinstance(v, collections.Mapping):
            base_settings[k] = recursive_dict_update(base_settings.get(k, {}), v)
        else:
            base_settings[k] = new_settings[k]
    return base_settings


def strip_strings(settings):
    """Recursively strip all strings in a settings dict.

    :param settings: the settings dictionary to process
    :returns: a new dictionary with all string settings stripped
    """
    result = {}
    for k, v in settings.items():
        if isinstance(v, str):
            result[k] = v.strip()
        elif isinstance(v, collections.Mapping):
            result[k] = strip_strings(v)
        else:
            result[k] = v
    return result



import abc
import warnings
from enum import Enum
from math import ceil

from .imaging import ImagingInformation, Modes


class LabViewVersions(Enum):
    pre2018 = "pre-2018 (original)"
    v231 = "2.3.1"


class LabViewHeader(metaclass=abc.ABCMeta):
    """A class to represent all information stored in a LabView header file.

    This abstract class only offers some basic functionality and structure.

    Different subclasses should be defined to provide special behaviour for
    particular versions of the LabView setup used in the lab.
    """
    property_names = {}

    @classmethod
    def from_file(cls, filename):
        """Create an object to hold the information in a LabView header file."""
        # Parse the file
        with open(filename, 'r') as ini:
            fields = []
            section = ''
            parsed_fields = {}
            for line in ini:
                line = line.strip()
                if len(line) > 0:
                    if line.startswith('['):
                        section = line[1:-1]
                        parsed_fields[section] = {}
                        line_number = 0  # used to count tab-separated lines
                    elif '=' in line:
                        words = line.split('=')
                        key, value = words[0].strip(), words[1].strip()
                        fields.append([section, key, value])
                        try:
                            value = float(value)
                        except ValueError:
                            pass
                        if isinstance(value, str) and value[0] == value[-1] == '"':
                            value = value[1:-1]
                        parsed_fields[section][key] = value
                    elif '\t' in line:
                        # Delay parsing these lines until later
                        line_number += 1
                        key = 'line_{}'.format(line_number)
                        fields.append([section, key, line])
                    else:
                        warnings.warn("Unrecognised non-blank line in {}: {}".format(filename, line))
        # Decide which version to instantiate
        try:
            version = parsed_fields['LOGIN']['Software Version']
        except KeyError:
            # older versions do not store the LabView version
            return LabViewHeaderPre2018(fields, parsed_fields)
        else:
            if version == '2.3.1':
                return LabViewHeader231(fields, parsed_fields)
            else:
                raise ValueError('Unsupported LabView version {}.'.format(version))

    def __init__(self, fields, processed_fields):
        """Create a header object from the given raw and processed fields.

        Subclasses can add specialised behaviour by overriding this.
        """
        # Keep track of the raw fields in case we need to return them later
        # (such as to add them to an NWB file).
        self._raw_fields = fields
        # Accessing the header's information should normally be through the
        # processed fields, stored in self.sections.
        self._sections = processed_fields
        self._imaging_mode = self._determine_imaging_mode()

    def __getitem__(self, item):
        """Retrieve an entry from the header's fields."""
        return self._sections[item]

    @property
    @abc.abstractmethod
    def version(self):
        """The version of LabView as found in the header file.

        Can be used to "branch" behaviour based on the version, rather than
        inspecting the class.
        """
        pass

    @property
    def imaging_mode(self):
        return self._imaging_mode

    @abc.abstractmethod
    def _determine_imaging_mode(self):
        pass

    @abc.abstractmethod
    def _imaging_section(self):
        """Get the section of the header that holds the imaging parameters."""
        pass

    def get_imaging_information(self):
        # Imaging parameters can be stored in different sections of the header
        # depending on the LabView version.
        section = self._imaging_section()
        # Parameter names also vary between versions, so use the appropriate ones.
        property_names = self.property_names
        # Also cast the integer parameters since they are read as floats.
        cycles_per_trial = int(section[property_names["number_of_cycles"]])
        gains = {"Red": section[property_names["gain_red"]],
                 "Green": section[property_names["gain_green"]]}
        frame_size = int(section[property_names["frame_size"]])
        field_of_view = section[property_names["field_of_view"]]
        number_of_miniscans = int(section[property_names["number_of_miniscans"]])
        dwell_time = section[property_names["dwell_time"]]
        return ImagingInformation(cycles_per_trial, gains, frame_size, field_of_view,
                                  number_of_miniscans, dwell_time)

    def determine_trial_times(self):
        """Try to extract the start and stop time of each trial.

        Raise an error if this is impossible because the information is not
        included in the header.
        """
        raise NotImplementedError

    def get_raw_fields(self):
        """Get the fields of the header as directly read from the file.

        Returns a list of (section, key, value) triples containing strings,
        without any processing applied. Useful for storing the original header
        in NWB files for clearer provenance.
        """
        return self._raw_fields


class LabViewHeaderPre2018(LabViewHeader):

    property_names = {
        "frame_size": "frame size",
        "field_of_view": "field of view",
        "dwell_time": "dwelltime (us)",
        "number_of_cycles": "number of cycles",
        "number_of_miniscans": "number of miniscans",
        "gain_red": "pmt 1",
        "gain_green": "pmt 2",
    }

    @property
    def version(self):
        return LabViewVersions.pre2018

    def _determine_imaging_mode(self):
        if self['GLOBAL PARAMETERS']['number of poi'] > 0:
            return Modes.pointing
        elif self['GLOBAL PARAMETERS']['number of miniscans'] > 0:
            return Modes.miniscan
        else:
            raise ValueError('Unsupported imaging type: numbers of poi and miniscans are zero.')

    def _imaging_section(self):
        # In the older version, parameters were stored in the global section.
        return self["GLOBAL PARAMETERS"]


class LabViewHeader231(LabViewHeader):

    property_names = {
        "frame_size": "Frame Size",
        "field_of_view": "field of view",
        "dwell_time": "pixel dwell time (us)",
        "number_of_cycles": "Number of cycles",
        "number_of_miniscans": "Number of miniscans",
        "gain_red": "pmt 1",
        "gain_green": "pmt 2",
    }

    # In this version of LabView, the trial times are stored in their own
    # (misleadingly titled) section of the header.
    trial_times_section = 'Intertrial FIFO Times'

    def __init__(self, fields, processed_fields):
        super().__init__(fields, processed_fields)
        self._parse_trial_times()

    @property
    def version(self):
        return LabViewVersions.v231

    def _determine_imaging_mode(self):
        volume_imaging = self['IMAGING MODES']['Volume Imaging']
        functional_imaging = self['IMAGING MODES']['Functional Imaging']
        if volume_imaging == 'TRUE' and functional_imaging == 'TRUE':
            raise ValueError('Unsupported imaging type: only one of "Volume '
                             'Imaging" and "Functional Imaging" can be true.')
        if volume_imaging == 'TRUE':
            return Modes.volume
        elif functional_imaging == 'TRUE':
            mode_name = self['FUNCTIONAL IMAGING']['Functional Mode']
            if mode_name == "Point":
                return Modes.pointing
            elif mode_name == "Patch":
                return Modes.miniscan
            else:
                raise ValueError('Unrecognised imaging mode: {}. Valid options'
                                 ' are: "Point", "Patch'.format(mode_name))
        else:
            raise ValueError('Unsupported imaging type: either "Volume Imaging"'
                             ' or "Functional Imaging" must be true.')

    def _imaging_section(self):
        # In LabView version 2.3.1, imaging parameters are stored under the
        # relevant imaging mode section.
        imaging_section_name = ("VOLUME IMAGING"
                                if self.imaging_mode is Modes.volume
                                else "FUNCTIONAL IMAGING")
        return self[imaging_section_name]

    def _parse_trial_times(self):
        # The relevant entries in the raw fields are triples whose first
        # element is the section header; select them based on that.
        time_fields = [field
                       for field in self._raw_fields
                       if field[0] == self.trial_times_section]
        assert len(time_fields) > 0, 'Trial times not found in header!'
        for line in time_fields:
            # The actual text of the line is the third element in the triple
            words = line[2].split('\t')
            assert len(words) == 2, 'Too many columns found for trial time'
            # Lines start with a line number (with decimal points) followed by a time
            key, value = int(float(words[0])), float(words[1])
            self._sections[self.trial_times_section][key] = value

    def determine_trial_times(self):
        trial_times = []
        parsed_times = self[self.trial_times_section]
        number_of_trials = ceil(len(parsed_times) / 2)
        # occasionally, the end time of the last trial will be missing,
        # in which case it will be assigned None and determined later from the speed ata
        last_time_present = (len(parsed_times) == 2 * number_of_trials)
        for i in range(number_of_trials):
            start = parsed_times[2 * i]
            if i < (number_of_trials-1) or last_time_present:
                end = parsed_times[2 * i + 1]
            else:
                end = None  # determine final 'end' later from speed data
            trial_times.append((start, end))
        return trial_times



import os

import pytest

from silverlabnwb import NwbFile
from silverlabnwb.signature import SignatureGenerator


def test_ambiguous_user_fails(tmpdir, ref_data_dir):
    fname = "test_metadata_import.nwb"
    nwb_path = os.path.join(str(tmpdir), fname)
    meta_path = os.path.join(ref_data_dir, 'meta_two_users.yaml')
    with pytest.raises(ValueError, match="Multiple users found in file."):
        with NwbFile(nwb_path, 'w') as nwb:
            nwb.create_from_metadata(meta_path)


def test_user_not_found_fails(tmpdir, ref_data_dir):
    fname = "test_metadata_import.nwb"
    nwb_path = os.path.join(str(tmpdir), fname)
    meta_path = os.path.join(ref_data_dir, 'meta_two_users.yaml')
    with pytest.raises(ValueError) as exc_info:
        with NwbFile(nwb_path, 'w') as nwb:
            nwb.create_from_metadata(meta_path, user="Mr Faker")
    assert "No session information found for user" in str(exc_info.value)


def test_no_start_time_fails(tmpdir, ref_data_dir):
    fname = "test_metadata_import.nwb"
    nwb_path = os.path.join(str(tmpdir), fname)
    meta_path = os.path.join(ref_data_dir, 'meta_two_users.yaml')
    with pytest.raises(ValueError) as exc_info:
        with NwbFile(nwb_path, 'w') as nwb:
            nwb.create_from_metadata(meta_path, user="A")
    assert "Start time for session not found!" == str(exc_info.value)


def test_metadata_import_correct(tmpdir, ref_data_dir):
    fname = "metadata_only_B.nwb"
    nwb_path = os.path.join(str(tmpdir), fname)
    meta_path = os.path.join(ref_data_dir, 'meta_two_users.yaml')
    sig_path = os.path.join(ref_data_dir, 'metadata_only_B.sig2')
    with NwbFile(nwb_path, 'w') as nwb:
        nwb.create_from_metadata(meta_path, user="B")
    sig_gen = SignatureGenerator()
    if os.environ.get('SILVERLAB_GEN_REF', '0') != '0':
        sig_gen.save_sig(nwb_path, sig_path)
    assert sig_gen.compare_to_sig(nwb_path, sig_path)



import os

import pytest

from silverlabnwb import NwbFile

DATA_PATH = os.environ.get('SILVERLAB_DATA_DIR', '')


@pytest.mark.skipif(
    not os.path.isdir(DATA_PATH),
    reason="raw data folder '{}' not present".format(DATA_PATH))
def test_all_funct_acq_writing():
    print("Searching for files")
    for subdirs in os.walk(DATA_PATH):
        import_folder_name = subdirs[0].split("\\")[-1]
        print("... in " + import_folder_name)
        if import_folder_name.startswith('sample') and not import_folder_name.endswith('VidRec'):
        # if import_folder_name.endswith('21 FunctAcq'):
            print("writing " + import_folder_name)
            with NwbFile(DATA_PATH + "\\nwb2\\" + import_folder_name.split(' ')[0] + ".nwb", mode='w') as nwb:
                nwb.import_labview_folder(subdirs[0])

    print("done with all.")


if __name__ == '__main__':
    print("Looking at folders in "+DATA_PATH)
    test_all_funct_acq_writing()



from enum import Enum


class Modes(Enum):
    """Scanning modes supported by the AOL microscope."""
    pointing = 1
    miniscan = 2
    patch = 2
    volume = 3


class ImagingInformation:
    """A class to hold imaging-related information found in the LabView header."""
    def __init__(self, cycles_per_trial, gains, frame_size, field_of_view,
                 number_of_miniscans, dwell_time):
        self.cycles_per_trial = cycles_per_trial
        self.gains = gains
        self.frame_size = frame_size
        self.field_of_view = field_of_view
        self.number_of_miniscans = number_of_miniscans
        self.dwell_time = dwell_time




import os

import h5py
import pytest

from silverlabnwb import NwbFile
from silverlabnwb.signature import SignatureGenerator

# Where to look for large raw data files
DATA_PATH = os.environ.get('SILVERLAB_DATA_DIR', '')


@pytest.fixture
def sig_gen():
    """A customised signature generator for large import tests."""
    sig_gen = SignatureGenerator()
    # The relative path changes on different systems
    sig_gen.ignore_path('/acquisition/.*/external_file')
    # This can be int32 or int64 depending on the Python version or operating system
    # but the schema defines its value as always 1.
    sig_gen.ignore_attribute('/(acquisition|stimulus)/.*/timestamps', 'interval')
    return sig_gen


@pytest.mark.skipif(
    os.environ.get('SILVERLAB_GEN_REF', '0') == '0',
    reason="SILVERLAB_GEN_REF not set or set to 0")
@pytest.mark.parametrize("nwb_name", [
    # Cut-down samples
    ('sample_pointing_videos_161215_15_34_21'),
    ('sample_pointing_fred_170317_10_11_01'),
    ('sample_miniscan_fred_170322_14_06_43'),
    ('sample_rois_200206_16_30_32'),
    # Full datasets
    ('161215_15_58_52'),
    ('161215_15_34_21'),
    ('170317_10_11_01'),
    ('170322_14_06_43'),
    ('161222_16_18_47'),
    ('170714_22_26_27')
])
def test_generate_signatures(ref_data_dir, sig_gen, nwb_name, monkeypatch):
    """A 'test' to generate reference data for the tests below."""
    sig_path = os.path.join(ref_data_dir, nwb_name + '.sig2')
    if os.environ.get('SILVERLAB_REGEN_NWB', '0') != '0':
        if nwb_name == 'sample_rois_200206_16_30_32':
            # Signature for sample_rois_200206_16_30_32 requires disabling _write_roi_data, as it contains no image data
            def do_nothing(*args):
                pass
            monkeypatch.setattr(NwbFile, "_write_roi_data", do_nothing)
        nwb_path = os.path.join(DATA_PATH, 'nwb2', nwb_name + '.nwb')
        labview_path = os.path.join(DATA_PATH,
                                    nwb_name + ' FunctAcq' if nwb_name[0] == '1' else nwb_name)
        with NwbFile(nwb_path, mode='w') as nwb:
            nwb.import_labview_folder(labview_path)
    else:
        nwb_path = os.path.join(DATA_PATH, nwb_name + '.nwb')
    sig_gen.save_sig(nwb_path, sig_path)


@pytest.fixture
def do_import_test(tmpdir, ref_data_dir, sig_gen):
    def do_import_test(expt, add_suffix=False):
        """Helper method for tests below."""
        nwb_path = os.path.join(str(tmpdir), expt + '.nwb')
        labview_path = os.path.join(DATA_PATH, expt + ' FunctAcq' if add_suffix else expt)
        sig_path = os.path.join(ref_data_dir, expt + '.sig2')

        with NwbFile(nwb_path, mode='w') as nwb:
            nwb.import_labview_folder(labview_path)
        assert sig_gen.compare_to_sig(nwb_path, sig_path)
    return do_import_test


@pytest.mark.skipif(
    not os.path.isdir(DATA_PATH),
    reason="raw data folder '{}' not present".format(DATA_PATH))
class TestSampleImports(object):

    def test_hana_video(self, do_import_test):
        do_import_test('sample_pointing_videos_161215_15_34_21')

    def test_fred_pointing(self, do_import_test):
        do_import_test('sample_pointing_fred_170317_10_11_01')

    def test_fred_miniscan(self, do_import_test):
        do_import_test('sample_miniscan_fred_170322_14_06_43')

    def test_harsha_rois(self, do_import_test, monkeypatch):
        # For now we don't include any imaging data for this, so disable
        # that part of the code
        def do_nothing(*args):
            pass
        monkeypatch.setattr(NwbFile, "_write_roi_data", do_nothing)
        do_import_test('sample_rois_200206_16_30_32')


@pytest.mark.skipif(
    not os.path.isdir(DATA_PATH),
    reason="raw data folder '{}' not present".format(DATA_PATH))
@pytest.mark.skipif(
    os.environ.get('SILVERLAB_TEST_LONG_IMPORTS', '0') == '0',
    reason="SILVERLAB_TEST_LONG_IMPORTS not set or set to 0")
class TestFullImporting(object):

    def test_hana(self, do_import_test):
        """A sample dataset from Hana with no videos."""
        do_import_test('161215_15_58_52', True)

    def test_hana_video(self, do_import_test):
        """A sample dataset from Hana with videos."""
        do_import_test('161215_15_34_21', True)

    def test_fred_pointing(self, do_import_test):
        """A sample dataset from Fred with pointing mode data."""
        do_import_test('170317_10_11_01', True)

    def test_fred_patch(self, do_import_test):
        """A sample dataset from Fred with miniscans."""
        do_import_test('170322_14_06_43', True)


@pytest.mark.skipif(
    not os.path.isdir(DATA_PATH),
    reason="raw data folder '{}' not present".format(DATA_PATH))
def test_roundtrip(tmpdir, ref_data_dir):
    """A simple check that file written can be read back in.

    Currently this does not do anything besides ensure there are no failures.
    Reading could fail, for example, if the extension spec is not included
    in the NWB file by accident.
    """
    # Write one of the subsampled experiments to NWB
    expt = 'sample_pointing_fred_170317_10_11_01'
    nwb_path = os.path.join(str(tmpdir), expt + '.nwb')
    labview_path = os.path.join(DATA_PATH, expt)
    with NwbFile(nwb_path, mode='w') as nwb:
        nwb.import_labview_folder(labview_path)

    # You would think the below would be a good check, but actually not.
    # The extensions are already loaded (from writing the file) and I can't
    # find a way to remove them.
    # with pynwb.NWBHDF5IO(nwb_path, 'r', load_namespaces=True) as io:
    #     new_file = io.read()
    # Right now, just check that the file has been read without errors
    # assert isinstance(new_file, pynwb.NWBFile)

    # Since the above isn't a meaningful test, roughly check that we have
    # included the extension spec and custom medata in the file.
    with h5py.File(nwb_path, 'r') as new_file:
        assert 'silverlab_extended_schema' in new_file['specifications']
        assert 'silverlab_metadata' in new_file['general']
        assert 'silverlab_optophysiology' in new_file['general']



"""Unit style tests for reading from various LabView headers (.ini files)"""
import os

import pytest

from silverlabnwb.header import LabViewHeader, LabViewVersions, Modes


@pytest.fixture(scope="module")
def header(request, ref_data_dir):
    """Create header object from a LabView header file."""
    header_file = os.path.join(ref_data_dir, request.param)
    header_object = LabViewHeader.from_file(header_file)
    return header_object


class TestLabViewHeaders(object):
    synthetic_header_path_v231 = 'synthetic experiment Header v231.ini'
    synthetic_header_path_v231_no_last_time = 'synthetic experiment Header v231 no last time.ini'
    synthetic_header_path_pre2018 = 'Experiment Header.ini'
    header_with_unrecognised_line_path = 'unrecognised line Header.ini'
    real_life_header_path_v231_pointing = 'real life Experiment Header v231 pointing.ini'

    @pytest.mark.parametrize("header, expected_version",
                             [(synthetic_header_path_v231, LabViewVersions.v231),
                              (synthetic_header_path_pre2018, LabViewVersions.pre2018),
                              (real_life_header_path_v231_pointing, LabViewVersions.v231)],
                             indirect=["header"])
    def test_lab_view_version(self, header, expected_version):
        assert header.version == expected_version

    @pytest.mark.parametrize("header, expected_mode",
                             [(synthetic_header_path_v231, Modes.miniscan),
                              (synthetic_header_path_pre2018, Modes.pointing),
                              (real_life_header_path_v231_pointing, Modes.pointing)],
                             indirect=["header"])
    def test_imaging_mode(self, header, expected_mode):
        assert header.imaging_mode == expected_mode

    @pytest.mark.parametrize("header, expected_trial_times",
                             [(synthetic_header_path_v231, [(0.0, 12.345678), (12.567890, 23.456789)]),
                              (synthetic_header_path_v231_no_last_time, [(0.0, 12.345678), (12.567890, None)])],
                             indirect=["header"])
    def test_trial_times(self, header, expected_trial_times):
        assert header.determine_trial_times() == expected_trial_times

    @pytest.mark.parametrize("header, expected_number_of_trials",
                             [(synthetic_header_path_v231, 2),
                              (synthetic_header_path_v231_no_last_time, 2),
                              (real_life_header_path_v231_pointing, 29)],
                             indirect=["header"])
    def test_number_of_trials(self, header, expected_number_of_trials):
        assert len(header.determine_trial_times()) == expected_number_of_trials

    @pytest.mark.parametrize("header",
                             [synthetic_header_path_pre2018],
                             indirect=["header"])
    def test_pre2018_trial_times_raises_error(self, header):
        with pytest.raises(NotImplementedError):
            header.determine_trial_times()

    def test_unrecognised_line_causes_warning(self):
        with pytest.warns(UserWarning) as list_of_warnings:
            LabViewHeader.from_file(os.path.join("tests", "data", self.header_with_unrecognised_line_path))
        assert len(list_of_warnings) == 1
        assert str(list_of_warnings[0].message).startswith("Unrecognised non-blank line")



import collections
import tkinter as T
from tkinter import messagebox, ttk

from idlelib.tooltip import Hovertip
from ruamel.yaml import YAML

from . import metadata


def wrap_dict(metadata):
    """Convert a metadata dict to use Tk variables to wrap entries."""
    return {key: wrap_value(value) for key, value in metadata.items()}


def wrap_list(metadata):
    """Convert a list of metadata components to use Tk variables."""
    return [wrap_value(item) for item in metadata]


def wrap_value(value):
    """Wrap a value as a Tk variable, or nested structure thereof.

    String values will automatically be stripped.
    """
    if isinstance(value, collections.Mapping):
        wrapped = wrap_dict(value)
    elif isinstance(value, list):
        wrapped = wrap_list(value)
    elif isinstance(value, str):
        wrapped = T.StringVar()
        wrapped.set(value.strip())
    elif isinstance(value, float):
        wrapped = T.DoubleVar()
        wrapped.set(value)
    elif isinstance(value, int):
        wrapped = T.IntVar()
        wrapped.set(value)
    elif isinstance(value, bool):
        wrapped = T.BooleanVar()
        wrapped.set(value)
    elif isinstance(value, T.Variable):
        # Clone the particular type of tkinter variable
        wrapped = type(value)()
        value = value.get()
        if hasattr(value, 'strip'):
            value = value.strip()
        wrapped.set(value)
    elif value is None:
        wrapped = T.StringVar()
        wrapped.set('')
    else:
        raise ValueError('Unexpected metadata item {} of type {}'.format(value, type(value)))
    return wrapped


def add_yaml_representers(yaml_instance):
    """Add YAML representers that convert Tk variables to their Python contents."""
    def get_repr(kind):
        def representer(dumper, data):
            data = data.get()
            rep = getattr(dumper, 'represent_' + kind)
            return rep(data)
        return representer
    yaml_instance.representer.add_representer(T.StringVar, get_repr('str'))
    yaml_instance.representer.add_representer(T.DoubleVar, get_repr('float'))
    yaml_instance.representer.add_representer(T.IntVar, get_repr('int'))
    yaml_instance.representer.add_representer(T.BooleanVar, get_repr('bool'))


class MetadataEditor(ttk.Frame):
    """A simple Tkinter GUI to help researchers fill in experiment metadata."""
    def __init__(self, master=None):
        """Initialise the GUI."""
        ttk.Frame.__init__(self, master)
        # Load the merged metadata from file, and extract the template experiment
        user_metadata, self.comments = metadata.read_user_config()
        self.template_expt = user_metadata['experiments'].pop('template')
        self.template_expt_comments = self.comments['experiments'].pop('template')
        del user_metadata['devices']  # Until we support editing it
        self.metadata = wrap_dict(user_metadata)
        self.yaml_instance = YAML()
        add_yaml_representers(self.yaml_instance)
        # self.original_metadata = copy.deepcopy(self.metadata)
        # Action buttons
        # ttk.Button(self, text="REVERT", command=self.revert).grid(
        #     row=0, column=0, sticky=T.W, padx=5, pady=5)
        ttk.Button(self, text="SAVE", command=self.save).grid(
            row=0, column=1, sticky=T.W, padx=5, pady=5)
        ttk.Button(self, text="DONE", command=self.done).grid(
            row=0, column=2, sticky=T.W, padx=5, pady=5)
        # Content tabs
        tabs = self.tabs = ttk.Notebook(self, padding=2)
        self.session_tab = self.make_session_tab(tabs)
        self.people_tab = self.make_people_tab(tabs)
        self.expts_tab = self.make_expts_tab(tabs)
        tabs.enable_traversal()
        tabs.grid(row=1, column=0, columnspan=3)
        # Resize support - make the content section take it all
        self.rowconfigure(1, weight=1)
        self.columnconfigure('all', weight=1)
        tabs.rowconfigure('all', weight=1)
        tabs.columnconfigure('all', weight=1)
        self.pack(side="top", fill="both", expand=True)

    def save(self):
        """Save current settings to file.

        TODO: Give the user some indication that this has happened!
        """
        self.record_expt()  # The only part that doesn't sync by itself
        metadata.save_config_file(self.metadata, self.yaml_instance)

    def done(self):
        """Save settings and quit the editor."""
        self.save()
        self.master.destroy()

    # def revert(self):
    #     """Revert all metadata changes made by this editor."""
    #     self.metadata = copy.deepcopy(self.original_metadata)
    #     self.save()

    def make_people_tab(self, parent):
        """Build the people tab and add it to the parent."""
        frame = ttk.Frame(parent)
        # Choosing who to edit
        ttk.Label(frame, text='Edit person:').grid(row=0, column=0, sticky='e')
        person_list = ttk.Combobox(frame, state='readonly', width=10,
                                   values=sorted(self.metadata['people'].keys()))
        person_list.bind('<<ComboboxSelected>>',
                         lambda e: self.update_people_tab(person_list.get()))
        person_list.grid(row=0, column=1, sticky='w')
        # Section for adding new people
        ttk.Label(frame, text='New person:').grid(
            row=2, column=0, sticky='e')
        new_person_var = T.StringVar()
        ttk.Entry(frame, textvariable=new_person_var, width=10).grid(
            row=2, column=1, sticky='w')
        # TODO: Make hitting return in the Entry click the Add button
        # TODO: Make the Add button stick next to the Entry box
        ttk.Button(frame, text='Add',
                   command=lambda: self.add_person(new_person_var.get(),
                                                   person_list)).grid(
            row=2, column=2, sticky='w')
        # Fields to edit current person
        widgets = self.session_widgets = {}
        ttk.Label(frame, text='Full name:').grid(row=0, column=4, sticky='e')
        widgets['name'] = ttk.Entry(frame)
        widgets['name'].grid(row=0, column=5, sticky='w')
        ttk.Label(frame, text='ORCID:').grid(row=1, column=4, sticky='e')
        widgets['orcid'] = ttk.Entry(frame)
        widgets['orcid'].grid(row=1, column=5, sticky='w')
        ttk.Label(frame, text='SCOPUS ID:').grid(row=2, column=4, sticky='e')
        widgets['scopus_id'] = ttk.Entry(frame)
        widgets['scopus_id'].grid(row=2, column=5, sticky='w')
        if self.metadata['people']:
            person_list.set(person_list['values'][0])
            self.update_people_tab(person_list.get())
        # Resize support
        frame.rowconfigure('all', weight=0)
        frame.columnconfigure(0, weight=2)
        frame.columnconfigure(3, weight=1)
        frame.columnconfigure(6, weight=2)
        parent.add(frame, text='People')
        return frame

    def update_people_tab(self, person_id):
        """Associate the entry boxes with the given person's details."""
        assert person_id in self.metadata['people']
        person = self.metadata['people'][person_id]
        widgets = self.session_widgets
        for key in ['name', 'orcid', 'scopus_id']:
            if key not in person:
                # Optional item was not present before
                person[key] = T.StringVar()
            widgets[key]['textvariable'] = person[key]

    def add_person(self, person_id, person_list_box):
        """Add a new researcher."""
        # TODO: Update the 'edit person' dropdown to match person_id
        existing_people = person_list_box['values']
        if person_id in existing_people:
            messagebox.showerror(message='The person ID "{}" already exists'.format(person_id))
            return
        if not person_id.isalnum():
            messagebox.showerror(message='A person ID can only contain letters and numbers')
            return
        person_list_box['values'] = sorted(existing_people + (person_id,))
        self.session_researcher_box['values'] = person_list_box['values']
        name_var = T.StringVar()
        name_var.set(person_id)
        self.metadata['people'][person_id] = {'name': name_var}
        self.update_people_tab(person_id)

    def make_session_tab(self, parent):
        """Build a tab for setting session properties."""
        frame = ttk.Frame(parent)
        ttk.Label(frame, text='Description:').grid(row=2, column=0, sticky='e')
        desc = ttk.Entry(frame, width=70)
        desc.grid(row=2, column=1, sticky='w')

        ttk.Label(frame, text='Researcher:').grid(row=0, column=0, sticky='e')
        researcher = ttk.Combobox(frame, state='readonly',
                                  values=sorted(self.metadata['people'].keys()))
        researcher.bind('<<ComboboxSelected>>',
                        lambda e: self.update_session(person=researcher.get(),
                                                      expt=experiment,
                                                      desc=desc))
        researcher.grid(row=0, column=1, sticky='w')
        self.session_researcher_box = researcher

        ttk.Label(frame, text='Experiment:').grid(row=1, column=0, sticky='e')
        experiment = ttk.Combobox(frame, state='readonly',
                                  values=sorted(self.metadata['experiments'].keys()))
        experiment.bind('<<ComboboxSelected>>',
                        lambda e: self.update_session(person=researcher.get(),
                                                      expt=experiment.get()))
        experiment.grid(row=1, column=1, sticky='w')
        self.session_experiment_box = experiment

        if 'last_session' not in self.metadata:
            self.metadata['last_session'] = T.StringVar()
            self.metadata['last_session'].set(researcher['values'][0])
        last_person = self.metadata['last_session'].get()
        researcher.set(last_person)
        self.update_session(person=researcher.get(), desc=desc, expt=experiment)
        # Resize support & add to tabs
        frame.rowconfigure('all', weight=0)
        frame.columnconfigure('all', weight=1)
        parent.add(frame, text='This session')
        return frame

    def update_session(self, person, desc=None, expt=None):
        """Update session info to reflect changes in the dropdowns."""
        self.metadata['last_session'].set(person)
        if person not in self.metadata['sessions']:
            self.metadata['sessions'][person] = {
                'description': T.StringVar(),
                'experiment': T.StringVar()
            }
            self.metadata['sessions'][person]['description'].set(
                "One or two sentences describing the experiment and data in the file.")
        if desc is not None:
            desc['textvariable'] = self.metadata['sessions'][person]['description']
        if expt is not None:
            expt_var = self.metadata['sessions'][person]['experiment']
            if isinstance(expt, str):
                expt_var.set(expt)
            else:
                expt.set(expt_var.get())

    def make_expts_tab(self, parent):
        """Build a tab for setting experiment info."""
        expt_frame = ttk.Frame(parent)
        # Control pane along the top
        frame = ttk.Frame(expt_frame)

        ttk.Label(frame, text='Edit experiment:').grid(row=0, column=0, sticky='e')
        experiment = ttk.Combobox(frame, state='readonly', width=15,
                                  values=sorted(self.metadata['experiments'].keys()))
        experiment.bind('<<ComboboxSelected>>',
                        lambda e: self.update_expts_tab(expt_id=experiment.get()))
        experiment.grid(row=0, column=1, sticky='w')
        self.experiment_shown = None
        self.experiment_selector = experiment

        ttk.Label(frame, text='Clone experiment as:').grid(row=1, column=0, sticky='e')
        clone_var = T.StringVar()
        ttk.Entry(frame, textvariable=clone_var, width=15).grid(row=1, column=1, sticky='we')
        self.clone_expt_button = ttk.Button(
            frame, text='Clone',
            command=lambda: self.add_expt(clone_var.get(),
                                          clone=experiment.get()))
        self.clone_expt_button.grid(row=1, column=2, sticky='w')
        if not self.metadata['experiments']:
            self.clone_expt_button.state(['disabled'])

        ttk.Label(frame, text='Add experiment:').grid(row=2, column=0, sticky='e')
        add_var = T.StringVar()
        ttk.Entry(frame, textvariable=add_var, width=15).grid(row=2, column=1, sticky='we')
        ttk.Button(frame, text='Add',
                   command=lambda: self.add_expt(add_var.get())).grid(
            row=2, column=2, sticky='w')

        frame.columnconfigure(0, weight=1)  # Expand at the sides
        frame.columnconfigure(3, weight=1)
        frame.pack(side='top', fill='x', expand=True)
        # Experiment property editing below the control pane; frame per 'item',
        # all in a scrolling canvas
        self.expts_canvas = canvas = T.Canvas(expt_frame, borderwidth=0)
        canvas.pack(side='left', fill='both', expand=True)
        frame = ttk.Frame(canvas)
        scrollbar = ttk.Scrollbar(expt_frame, orient='vertical', command=canvas.yview)
        scrollbar.pack(side='right', fill='y')
        canvas.configure(yscrollcommand=scrollbar.set)
        canvas.create_window((0, 0), window=frame, anchor='nw')
        frame.bind("<Configure>", self.on_expts_frame_configure)
        # The contents
        self.expts_boxes = {}
        self.make_expts_part(frame, 'description')
        self.make_expts_part(frame, 'optophysiology')
        self.make_stim_frame(frame)
        for part in ['data_collection', 'pharmacology', 'protocol', 'slices', 'stimulus',
                     'subject', 'surgery', 'virus', 'related_publications', 'notes']:
            self.make_expts_part(frame, part)
        # Show current experiment, if any
        if self.metadata['experiments']:
            experiment.set(experiment['values'][0])
            self.update_expts_tab(expt_id=experiment.get())
        # Add to tabs
        parent.add(expt_frame, text='Experiments')
        return expt_frame

    def on_expts_frame_configure(self, event):
        """Reset the scroll region to encompass the inner frame."""
        self.expts_canvas.configure(scrollregion=self.expts_canvas.bbox("all"))

    def add_expt(self, expt_id, clone=None):
        """Create a new experiment and start editing it."""
        existing_expts = self.experiment_selector['values'] or ()
        if expt_id in existing_expts:
            messagebox.showerror(message='The experiment ID "{}" already exists'.format(expt_id))
            return
        if not expt_id or not expt_id[0].isalpha():
            messagebox.showerror(message='An experiment ID must start with a letter')
            return
        if clone is None:
            clone = self.template_expt
        else:
            self.record_expt()
            clone = self.metadata['experiments'][clone]
        self.metadata['experiments'][expt_id] = wrap_dict(clone)
        self.experiment_selector['values'] = sorted(existing_expts + (expt_id,))
        self.experiment_selector.set(expt_id)
        self.session_experiment_box['values'] = self.experiment_selector['values']
        self.update_expts_tab(expt_id)

    def update_expts_tab(self, expt_id):
        """Show the given experiment in the GUI for editing."""
        def do_subfields(box, expt_field):
            for subfield, entry in box.items():
                if isinstance(entry, collections.Mapping):
                    do_subfields(entry, expt_field[subfield])
                else:
                    entry['textvariable'] = expt_field[subfield]
        if expt_id == self.experiment_shown:
            return
        self.record_expt()
        expt = self.metadata['experiments'][expt_id]
        for field, box in self.expts_boxes.items():
            if isinstance(box, T.Text):
                box.delete('1.0', 'end')
                if field in expt:
                    box.insert('1.0', expt[field].get())
            elif isinstance(box, collections.Mapping):
                do_subfields(box, expt[field])
            elif isinstance(box, list):
                for i, item in enumerate(box):
                    do_subfields(item, expt[field][i])
        self.experiment_shown = expt_id
        self.clone_expt_button.state(['!disabled'])

    def record_expt(self):
        """Save the currently shown experiment fields as the selected experiment.

        We only need to worry about Text fields; Entry fields are tracked by their StringVar.
        Fields that are empty are omitted.
        """
        expt_id = self.experiment_shown
        if expt_id is None:
            return
        expt = self.metadata['experiments'][expt_id]
        for field, box in self.expts_boxes.items():
            if isinstance(box, T.Text):
                contents = box.get('1.0', 'end').strip()
                if contents:
                    expt[field] = T.StringVar()
                    expt[field].set(contents)
                elif field in expt:
                    del expt[field]

    def make_label(self, parent, name, row=0, sticky='w', tooltip_text=''):
        """Make the human-friendly label for a form section."""
        name = name.capitalize().replace('_', ' ') + ':'
        label = ttk.Label(parent, text=name)
        label.grid(row=row, column=0, sticky=sticky)
        if len(tooltip_text) > 0:
            Hovertip(label, tooltip_text, hover_delay=1000)  # delay in ms

    def make_expts_part(self, parent, part_name):
        """Make a component frame for the experiment editor."""
        frame = ttk.Frame(parent)
        if isinstance(self.template_expt[part_name], collections.Mapping):
            # This is actually a related group of fields
            self.make_label(frame, part_name)
            self.expts_boxes[part_name] = boxes = {}
            self.make_expts_fields(frame, self.template_expt[part_name], boxes,
                                   self.template_expt_comments[part_name])
        else:
            self.make_label(frame, part_name, tooltip_text=self.template_expt_comments[part_name])
            textbox = T.Text(frame, width=100, height=5, wrap='word')
            textbox.grid(row=1, column=0, sticky='nesw')
            expts_part = self.template_expt[part_name]
            textbox.insert('1.0', expts_part if expts_part is not None else "")
            self.expts_boxes[part_name] = textbox
        frame.pack(side='top', fill='x', expand=True)

    def make_expts_fields(self, frame, template, boxes, comments):
        """Create a related group of fields for an experiment section."""
        for i, field in enumerate(template):
            if isinstance(template[field], collections.Mapping):
                self.make_label(frame, field, row=i + 1, sticky='ne')
                # Another nesting level!
                boxes[field] = {}
                subframe = ttk.Frame(frame)
                for j, subfield in enumerate(template[field]):
                    self.make_label(subframe, subfield, row=j, sticky='e',
                                    tooltip_text=comments[field][subfield])
                    boxes[field][subfield] = ttk.Entry(subframe, width=40)
                    boxes[field][subfield].grid(row=j, column=1, sticky='w')
                subframe.grid(row=i + 1, column=1, sticky='we')
            else:
                tooltip = comments.get(field, '')
                self.make_label(frame, field, row=i + 1, sticky='ne',
                                tooltip_text=tooltip)
                boxes[field] = ttk.Entry(frame, width=70)
                boxes[field].grid(row=i + 1, column=1, sticky='w')

    def make_stim_frame(self, parent):
        """Create fields to edit the stimulus details.

        TODO: At present this restricts you to the same number of stimuli as the template.
        """
        frame = ttk.Frame(parent)
        part_name = 'stimulus_details'
        self.make_label(frame, part_name)
        self.expts_boxes[part_name] = []
        for i, stim in enumerate(self.template_expt[part_name]):
            boxes = {}
            self.expts_boxes[part_name].append(boxes)
            self.make_expts_fields(frame, stim, boxes,
                                   self.template_expt_comments[part_name][i])
        frame.pack(side='top', fill='x', expand=True)


def run_editor():
    '''Run the metadata editor.'''
    app = MetadataEditor()
    app.master.title("Silver Lab NWB Metadata Editor")
    app.lift()  # Raise to top of window stack
    app.mainloop()



#!/usr/bin/env python

"""A simple script to generate test datasets by subsampling.

Given a Labview export folder and previously imported NWB file,
it will generate a new 'export' folder with a subset of the ROIs
and trials from the original, and compressed videos/images.
"""


import argparse
import glob
import os
import subprocess

import av
import h5py
import nptdms
import numpy as np
from PIL import Image


def subsample_nwb(nwb, input_path, output_path, ntrials=2, nrois=10):
    """Create a subset of an NWB Labview structure for testing purposes.

    :param nwb: h5py.File for the input NWB
    :param output_path: folder to write output to
    :param ntrials: how many trials to include
    :param nrois: how many ROIs to include
    """
    print('Processing', nwb.filename)
    os.makedirs(output_path, exist_ok=True)
    # Copy truncated ROI.dat
    orig_nrois = copy_roi_dat(input_path, output_path, nrois)
    # Copy modified header
    copy_header(
        input_path, output_path,
        ntrials, nrois, orig_nrois)
    # Figure out time duration for given ntrials
    last_trial = nwb['/epochs/trial_{:04d}'.format(ntrials)]
    end_time = last_trial['stop_time'].value
    print('Trial {} ends at {}'.format(ntrials, end_time))
    # Copy truncated speed data
    copy_speed_data(input_path, output_path, last_trial)
    # Figure out which Zstack planes have ROIs
    zstack_planes = find_used_planes(nwb, nrois)
    # Copy pockels file
    copy_pockels_file(input_path, output_path, zstack_planes)
    # Copy & compress Zstack images
    copy_zstack(input_path, output_path, zstack_planes)
    # Copy subset of TDMS files
    tdms_path = 'Functional imaging TDMS data files'
    os.makedirs(os.path.join(output_path, tdms_path), exist_ok=True)
    for tr in range(1, ntrials + 1):
        tdms_name = '{:03d}.tdms'.format(tr)
        tdms_in = os.path.join(input_path, tdms_path, tdms_name)
        tdms_out = os.path.join(output_path, tdms_path, tdms_name)
        copy_tdms(nwb, tdms_in, tdms_out, nrois)
    # Find videos defined
    video_names = [name for name in nwb['/acquisition/timeseries'].keys()
                   if name.endswith('Cam')]
    print('Videos:', video_names)
    # Compress videos
    if video_names:
        video_output_path = output_path + ' VidRec'
        os.makedirs(video_output_path, exist_ok=True)
    for name in video_names:
        ts = nwb['/acquisition/timeseries/' + name]
        old_path = ts['external_file'][0].decode()
        new_path = os.path.join(video_output_path, os.path.basename(old_path))
        times = ts['timestamps'].value
        frames_to_keep = np.count_nonzero(times <= end_time)
        compress_video(old_path, new_path, frames_to_keep)


def find_used_planes(nwb, nrois):
    used_planes = set()
    seg_iface = nwb['/processing/Acquired_ROIs/ImageSegmentation']
    for plane_name in seg_iface.keys():
        plane_num = int(plane_name[-4:])
        for roi_name in seg_iface[plane_name].keys():
            if roi_name.startswith('ROI_'):
                roi_num = int(roi_name[4:])
                if roi_num <= nrois:
                    used_planes.add(plane_num)
    used_planes = list(used_planes)
    used_planes.sort()
    print('Relevant Zstack planes are:', used_planes)
    return used_planes


def copy_zstack(input_path, output_path, zstack_planes):
    dirname = 'Zstack Images'
    sources = glob.glob(os.path.join(input_path, dirname, '*Channel*.tif'))
    os.makedirs(os.path.join(output_path, dirname), exist_ok=True)
    for source in sources:
        dest = os.path.join(output_path, os.path.relpath(source, input_path))
        fname = os.path.basename(source)
        stack_index = int(fname.split('_')[1][:-4])
        if stack_index in zstack_planes:
            print('Copying frame {}'.format(fname))
            with Image.open(source) as im:
                im.save(dest, format='TIFF', compression='tiff_lzw')


def copy_speed_data(input_path, output_path, last_trial):
    fname = 'Speed_Data/Speed data 001.txt'
    os.makedirs(os.path.join(output_path, 'Speed_Data'), exist_ok=True)
    src = os.path.join(input_path, fname)
    dest = os.path.join(output_path, fname)
    speed_data_ts = last_trial['speed_data']
    end_index = speed_data_ts['idx_start'].value + speed_data_ts['count'].value
    copy_and_truncate(src, dest, end_index + 1)


def copy_pockels_file(input_path, output_path, zstack_planes):
    fname = 'Zplane_Pockels_Values.dat'
    with open(os.path.join(output_path, fname), 'wt') as outf:
        with open(os.path.join(input_path, fname), 'rt') as inf:
            plane_index = -5
            for line in inf:
                if plane_index < 1 or plane_index in zstack_planes:
                    outf.write(line)
                # plane_index += 1


def copy_header(input_path, output_path, ntrials, nrois, orig_nrois):
    def get_value(line):
        return float(line.split('=')[1].strip())
    roi_marker = 'number of poi'
    miniscan_marker = 'number of miniscans'
    trial_marker = 'number of trials'
    fname = 'Experiment Header.ini'
    src = os.path.join(input_path, fname)
    dest = os.path.join(output_path, fname)
    with open(dest, 'wt') as outf:
        with open(src, 'rt') as inf:
            for line in inf:
                if line.startswith(roi_marker):
                    if get_value(line) > 0:
                        # Pointing mode experiment
                        line = '{} = {}\n'.format(roi_marker, nrois)
                        new_nscans = nrois
                elif line.startswith(miniscan_marker):
                    n_scans = get_value(line)
                    if n_scans > 0:
                        # Patch mode experiment
                        new_nscans = n_scans / orig_nrois * nrois
                        line = '{} = {}\n'.format(miniscan_marker, new_nscans)
                elif line.startswith(trial_marker):
                    line = '{} = {}\n'.format(trial_marker, ntrials)
                outf.write(line)
    # Only `new_nscans` lines of cycle times are now needed
    cycle_fname = 'Single cycle relative times.txt'
    copy_and_truncate(os.path.join(input_path, cycle_fname),
                      os.path.join(output_path, cycle_fname),
                      int(new_nscans))


def copy_roi_dat(input_path, output_path, nrois):
    filename = 'ROI.dat'
    src = os.path.join(input_path, filename)
    dest = os.path.join(output_path, filename)
    copy_and_truncate(src, dest, nrois + 1)
    with open(dest, 'wt') as outf:
        with open(src, 'rt') as inf:
            for i in range(nrois + 1):
                outf.write(inf.readline())
            orig_nrois = nrois
            for line in inf:
                if line.strip():
                    orig_nrois += 1
    print('Retaining {} ROIs of {}'.format(nrois, orig_nrois))
    return orig_nrois


def copy_and_truncate(src, dest, nlines):
    print('Copying {} lines of {} to {}'.format(nlines, src, dest))
    with open(dest, 'wt') as outf:
        with open(src, 'rt') as inf:
            for i in range(nlines):
                outf.write(inf.readline())


def cycles_per_trial(nwb):
    """Get the number of microscope cycles/trial.

    That is, the number of times each point is imaged in each
    trial. Currently looks at the first imaging timeseries in
    the first trial, and assumes they're all the same.
    """
    trial1 = nwb['/epochs/trial_0001']
    for ts_name in trial1:
        ts = trial1[ts_name]
        is_image_series = ts['timeseries/pixel_time_offsets'] is not None
        if is_image_series:
            return ts['count'].value
    else:
        raise ValueError('No imaging timeseries found')


def copy_tdms(nwb, in_path, out_path, nrois):
    num_all_rois = nwb['/processing/Acquired_ROIs/roi_spec'].shape[0]
    print('Copying {} of {} ROIs from {} to {}'.format(
        nrois, num_all_rois, in_path, out_path))
    in_tdms = nptdms.TdmsFile(in_path)
    group_name = 'Functional Imaging Data'
    with nptdms.TdmsWriter(out_path) as out_tdms:
        root, group = in_tdms.object(), in_tdms.object(group_name)
        out_tdms.write_segment([root, group])
        for ch, channel in {'0': 'Red', '1': 'Green'}.items():
            ch_name = 'Channel {} Data'.format(ch)
            ch_obj = in_tdms.object(group_name, ch_name)
            shape = (cycles_per_trial(nwb), num_all_rois, -1)
            ch_data = ch_obj.data.reshape(shape)
            subset = ch_data[:, :nrois, :].reshape(-1)
            new_obj = nptdms.ChannelObject(group_name, ch_name, subset, properties={})
            out_tdms.write_segment([new_obj])


def compress_video(input_path, output_path, nframes, width=300):
    print('Compressing {} frames of "{}" to "{}"'.format(
        nframes, input_path, output_path))
    input_file = av.open(input_path)
    input_vid = input_file.streams.video[0]
    assert input_vid.frames >= nframes, 'Only {} frames present in video'.format(
        input_vid.frames)
    scale = input_vid.width / width
    height = int(round(input_vid.height / scale))
    if height % 2 == 1:
        height += 1
    print('Input {}x{} @ {}, {}'.format(
        input_vid.width, input_vid.height, input_vid.rate, input_vid.bit_rate))

    # Create a ffmpeg command line for the compression & resizing
    args = [
        'ffmpeg',
        '-i', input_path,
        '-y',
        '-frames:v', str(nframes),
        '-s', '{}x{}'.format(width, height),
        '-pix_fmt', 'yuvj420p',
        '-codec:v', 'mjpeg',
        output_path
    ]
    subprocess.check_call(args)

    # Copy truncated 'relative times' file
    base = os.path.basename(input_path).split('-')[0]
    times_name = base + '-relative times.txt'
    src = os.path.join(os.path.dirname(input_path), times_name)
    dest = os.path.join(os.path.dirname(output_path), times_name)
    copy_and_truncate(src, dest, nframes)


def create_parser():
    parser = argparse.ArgumentParser(
        description='Prepare subsampled test datasets for NWB import')
    parser.add_argument('input_id',
                        help='experiment ID to subsample')
    parser.add_argument('output_path',
                        help='output folder path to write')
    parser.add_argument('--nrois', '-r',
                        type=int, default=10,
                        help='number of ROIs to retain')
    parser.add_argument('--ntrials', '-t',
                        type=int, default=2,
                        help='number of trials to retain')
    return parser


def run():
    args = create_parser().parse_args()
    input_path = args.input_id + ' FunctAcq'
    nwb_path = args.input_id + '.nwb'
    with h5py.File(nwb_path, mode='r') as nwb:
        subsample_nwb(nwb, input_path, args.output_path,
                      nrois=args.nrois, ntrials=args.ntrials)


if __name__ == '__main__':
    run()



import os

import pytest

from silverlabnwb import timings


@pytest.fixture(scope="module")
def synthetic_timings_v231(ref_data_dir):
    timings_file_name = "synthetic v231 Single cycle relative times_HW.txt"
    timings_file_path = os.path.join(ref_data_dir, timings_file_name)
    roi_file_name = "synthetic v231 ROI.dat"
    roi_file_path = os.path.join(ref_data_dir, roi_file_name)
    # the synthetic file has
    # 2 trials, each of which has 3 cycles and 4 rois a 5 lines
    # and a few zero lines as may be expected "in the wild"
    # first cycle of trial 1 and trial 2 take 1300.4 and 1200.4 nanoseconds, respectively
    return timings.LabViewTimings231(relative_times_path=timings_file_path,
                                     roi_path=roi_file_path,
                                     n_cycles_per_trial=3,
                                     n_trials=2,
                                     dwell_time=1.e-6)


@pytest.fixture(scope="module")
def synthetic_timings_pre2018(ref_data_dir):
    timings_file_name = "synthetic pre2018 Single cycle relative times.txt"
    timings_file_path = os.path.join(ref_data_dir, timings_file_name)
    roi_file_name = "synthetic pre2018 ROI.dat"
    roi_file_path = os.path.join(ref_data_dir, roi_file_name)
    return timings.LabViewTimingsPre2018(relative_times_path=timings_file_path,
                                         roi_path=roi_file_path,
                                         dwell_time=1.e-6)


def test_cycle_time_v231(synthetic_timings_v231):
    expected_mean_first_cycle_time_s = 1250.4 / 1e6
    assert synthetic_timings_v231.cycle_time == expected_mean_first_cycle_time_s


@pytest.mark.parametrize("roi_index", list(range(4)))  # 4 ROIs
def test_pixel_time_offsets_for_rois_v231(synthetic_timings_v231, roi_index):
    roi_offsets = synthetic_timings_v231.pixel_time_offsets[roi_index]
    expected_shape = (6, 5, 6)  # 2 trials * 3 cycles , 5 lines/roi, 6 pixels/line
    expected_first_cycle_first_row_offset = (200 + 300 * roi_index) / 1e6
    expected_first_cycle_last_row_offset = (400.4 + 300 * roi_index) / 1e6
    expected_last_cycle_first_row_offset = (4100 + 300 * roi_index) / 1e6
    assert roi_offsets.shape == expected_shape
    assert roi_offsets[0][0][0] == expected_first_cycle_first_row_offset
    assert roi_offsets[0][4][0] == expected_first_cycle_last_row_offset
    assert roi_offsets[-1][0][0] == expected_last_cycle_first_row_offset


def test_cycle_time_pre2018(synthetic_timings_pre2018):
    expected_cycle_time = 12345 / 1e6
    assert synthetic_timings_pre2018.cycle_time == expected_cycle_time


@pytest.mark.parametrize("roi_index", list(range(2)))  # 2 ROIs
def test_pixel_time_offsets_for_roi_pre2018(synthetic_timings_pre2018, roi_index):
    roi_offsets = synthetic_timings_pre2018.pixel_time_offsets[roi_index]
    expected_shape = (4, 10)  # 4 lines/roi, 10 pixels/line
    expected_first_row_offset = (400 + 400 * roi_index) / 1e6
    expected_last_row_offset = (700 + 400 * roi_index) / 1e6
    assert roi_offsets.shape == expected_shape
    assert roi_offsets[0][0] == expected_first_row_offset
    assert roi_offsets[-1][0] == expected_last_row_offset



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
import io
import re
from glob import glob
from os.path import basename, dirname, join, splitext

from setuptools import find_packages, setup


def read(*names, **kwargs):
    with io.open(
        join(dirname(__file__), *names),
        encoding=kwargs.get('encoding', 'utf8')
    ) as fh:
        return fh.read()


setup(
    name='silverlabnwb',
    version='0.2.0',
    license='MIT license',
    description='Python tools for working with Silver Lab data in the NWB2 format',
    long_description='%s\n%s' % (
        re.compile('^.. start-badges.*^.. end-badges', re.M | re.S).sub('', read('README.rst')),
        re.sub(':[a-z]+:`~?(.*?)`', r'``\1``', read('CHANGELOG.rst'))
    ),
    author='UCL Research Software Development Group',
    author_email='rc-softdev@ucl.ac.uk',
    url='https://github.com/SilverLabUCL/PySilverLabNWB',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    py_modules=[splitext(basename(path))[0] for path in glob('src/*.py')],
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Operating System :: Unix',
        'Operating System :: POSIX',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython',
    ],
    keywords=[
        # eg: 'keyword1', 'keyword2', 'keyword3',
    ],
    install_requires=[
        'pynwb==1.3.2',
        'appdirs',
        'h5py>=2.7.1',
        'hdmf<2.0.0',
        'nptdms<=0.26.0',
        'numpy',
        'pandas>=0.20',
        'ruamel.yaml',
        'tifffile[all]',  # [all] needed in newer versions of tifffile to ensure imagecodecs is included.
    ],
    extras_require={
        'test': ['pytest', 'tox'],
        'video': ['av'],
    },
    entry_points={
        'console_scripts': [
            'labview2nwb = silverlabnwb.script:import_labview',
            'subsample_nwb = silverlabnwb.subsample_nwb:run',
            'nwb_sig = silverlabnwb.signature:cli',
            'nwb_sig_convert = silverlabnwb.signature:convert_sig_cli',
            'metadata2nwb = silverlabnwb.script:import_metadata'
        ],
        'gui_scripts': [
            'nwb_metadata_editor = silverlabnwb.metadata_gui:run_editor'
        ]
    },
)



"""Testing that epochs are calculated and added correctly."""

import os

import h5py
import numpy as np
import pandas as pd
from ruamel.yaml import YAML

from silverlabnwb import NwbFile
from silverlabnwb.nwb_file import LabViewVersions

ignored = ['labview_header']


def compare_hdf5(nwb_path, expected_yaml_path):
    """Test utility method comparing a generated NWB file against expected contents.

    As a side-effect, checks that we can open files for reading with our API.
    """
    global current_file
    with open(expected_yaml_path, 'r') as f:
        yaml_instance = YAML(typ='safe')
        expected = yaml_instance.load(f)
    with NwbFile(nwb_path, mode='r') as nwb:
        current_file = nwb.hdf_file
        compare_group(nwb.hdf_file, expected, '')
        current_file = None


def compare_group(nwb_group, expected_group, path):
    """Check that an HDF5 group has the expected contents."""
    for key in expected_group:
        if key in ignored:
            continue
        expected_value = expected_group[key]
        if key == '_attrs':
            # Check attributes of the node
            compare_attributes(nwb_group, expected_value, path)
        elif key == '_value':
            # nwb_group should actually be a dataset
            assert isinstance(nwb_group, h5py.Dataset)
            compare_dataset(nwb_group, expected_value, path)
        elif key == '_link':
            # This group should be a soft link to another
            link = nwb_group.get(nwb_group.name, getlink=True)
            assert isinstance(link, h5py.SoftLink)
            assert link.path == expected_value
        else:
            assert key in nwb_group
            if isinstance(expected_value, dict):
                if '_columns' in expected_value:
                    compare_table(nwb_group[key], expected_value['_columns'], path + '/' + key)
                else:
                    compare_group(nwb_group[key], expected_value, path + '/' + key)
            else:
                compare_dataset(nwb_group[key], expected_value, path + '/' + key)


def compare_attributes(nwb_node, expected_attrs, path):
    """Check that an HDF5 node has the expected attributes."""
    for attr_name, attr_value in expected_attrs.items():
        assert attr_name in nwb_node.attrs
        compare_dataset(nwb_node.attrs[attr_name], attr_value, path + '/@' + attr_name)


def compare_generic_dataset(nwb_dataset, expected_value, path):
    """Check that an HDF5 dataset has the expected contents.

    Note that this gets used for both 'normal' datasets and attribute values.
    In the former case we must access the numpy value with [()]; in the latter
    nwb_dataset is already the numpy value.
    """
    if isinstance(nwb_dataset, h5py.Dataset):
        # Extract the actual data from the dataset
        nwb_dataset = nwb_dataset[()]
    if isinstance(expected_value, str):
        if isinstance(nwb_dataset, np.bytes_):
            # Convert to string so we can compare naturally
            nwb_dataset = nwb_dataset.decode('UTF-8')
        assert nwb_dataset == expected_value, 'Mismatch at {}'.format(path)
    elif isinstance(expected_value, (int, float)):
        assert abs(nwb_dataset - expected_value) < 1e-6, 'Mismatch at {}'.format(path)
    elif isinstance(expected_value, list):
        expected_value = np.array(expected_value)
        assert nwb_dataset.shape == expected_value.shape, 'Mismatch at {}'.format(path)
        if expected_value.dtype.kind == 'U':
            expected_value = expected_value.astype('S')
        if nwb_dataset.dtype.kind in ['O', 'U']:
            nwb_dataset = nwb_dataset.astype('S')
        if expected_value.dtype.kind == 'S':
            np.testing.assert_array_equal(nwb_dataset, expected_value)
        else:
            np.testing.assert_allclose(nwb_dataset, expected_value, atol=1e-6)
    elif isinstance(expected_value, dict):
        if '_columns' in expected_value:
            compare_table(nwb_dataset, expected_value['_columns'], path)
        else:
            compare_references(nwb_dataset, expected_value, path)
    else:
        assert 0, 'Unexpected expected_value {!r}'.format(expected_value)


def compare_references(nwb_dataset, expected_value, path):
    """Compare an array of object/region references.

    It's hard to check region references, as we can't pull out the offset directly.
    But we do what we can!
    """
    print('checking ref', path)
    assert len(nwb_dataset) == len(expected_value['_targets']), 'Wrong array length at ' + path
    assert len(nwb_dataset) > 0
    first_ref = nwb_dataset[0]
    assert type(first_ref).__name__ == expected_value['_type'], 'Wrong reference type at ' + path
    # Check all references point to the correct dataset/group
    for i, ref in enumerate(nwb_dataset):
        assert current_file[ref].name == expected_value['_targets'][i], \
            'Reference at {} points to {} not {}'.format(
                path, current_file[ref].name, expected_value['_targets'][i])
    # For region references, do some content checking
    if expected_value['_type'] == 'RegionReference':
        for i, ref in enumerate(nwb_dataset):
            target = current_file[ref][ref]
            print('checking rref', current_file[ref].name, i)
            assert target.shape == tuple(expected_value['_shapes'][i]), \
                'Wrong shape at {} ref {}'.format(path, i)
            compare_generic_dataset(target, expected_value['_values'][i], path + '@{}'.format(i))


def compare_table(nwb_dataset, expected_columns, path):
    """Compare a structured dataset against expected column values."""
    assert nwb_dataset.dtype.kind == 'V', 'Not a table at {}'.format(path)
    for colname in expected_columns:
        assert colname in nwb_dataset.dtype.names, 'Missing column {} at {}'.format(
            colname, path)
        if isinstance(nwb_dataset, h5py.Dataset):
            nwb_dataset = nwb_dataset[()]
        col = nwb_dataset[colname]
        compare_generic_dataset(col, expected_columns[colname], path + '#' + colname)


def compare_device(nwb_dataset, expected_value, path):
    """Check that a dataset representing a Device has the expected contents."""
    # Temporarily disabled due to not knowing where to put the description...
    # assert nwb_dataset.attrs['source'] == expected_value, 'Mismatch at {}'.format(path)
    pass


def compare_datetime(nwb_dataset, expected_value, path):
    """Check that a dataset containing a timestamp has the expected contents.

    This may be overkill, but it insulates us against different systems writing
    the date in different format."""
    assert pd.Timestamp(nwb_dataset[()].decode()) == pd.Timestamp(expected_value),\
        'Mismatch at {}'.format(path)


def compare_dataset(nwb_dataset, expected_value, path):
    """Check an HDF5 dataset, accounting for the type of its contents."""
    if '/devices/' in path:
        compare_device(nwb_dataset, expected_value, path)
    elif path.endswith('session_start_time'):
        compare_datetime(nwb_dataset, expected_value, path)
    else:
        compare_generic_dataset(nwb_dataset, expected_value, path)


def test_metadata_only(tmpdir, capfd, ref_data_dir):
    fname = "test_metadata.nwb"
    with NwbFile(os.path.join(str(tmpdir), fname), mode='w') as nwb:
        speed_data, start_time = nwb.create_nwb_file(ref_data_dir, 'test_metadata')
        nwb.add_core_metadata()
    compare_hdf5(str(tmpdir.join(fname)), os.path.join(ref_data_dir, 'expected_meta_only.yaml'))


def test_determine_labview_version(tmpdir, ref_data_dir):
    fname = "test_labview_version.nwb"
    with NwbFile(os.path.join(str(tmpdir), fname), mode='w') as nwb:
        speed_data, start_time = nwb.create_nwb_file(ref_data_dir, 'test_labview_version')
    assert nwb.labview_version is LabViewVersions.pre2018


def test_epochs(tmpdir, capfd, ref_data_dir):
    fname = "test_epochs.nwb"
    nwb_path = os.path.join(str(tmpdir), fname)
    with NwbFile(nwb_path, mode='w') as nwb:
        speed_data, start_time = nwb.create_nwb_file(ref_data_dir, 'test_epochs')
        nwb.add_core_metadata()
        nwb.add_speed_data(speed_data, start_time)
        nwb.determine_trial_times()
    compare_hdf5(nwb_path, os.path.join(ref_data_dir, 'expected_epochs.yaml'))



import abc

import numpy as np
import pandas as pd


class LabViewTimings(metaclass=abc.ABCMeta):
    """A class for storing pixel-level timing information for an ROI."""

    def __init__(self, relative_times_path, roi_path, dwell_time):
        self.dwell_time = dwell_time
        self._read_relative_times_file(relative_times_path)
        self._read_roi_data(roi_path)

    @abc.abstractmethod
    def _read_relative_times_file(self, file_path):
        pass

    def _read_roi_data(self, roi_path):
        self.roi_data = pd.read_csv(roi_path, sep='\t', dtype=np.float64)
        self.n_rois = len(self.roi_data['ROI index'])
        # in the future, we can access more shape parameters here too, for variable size ROIs in those cases,
        # we will need to access each individual row here separately. for now, take first row and all ROIs are same
        # size and orientation. Also, we may need to store the angle somewhere in the future (to reconstruct the ROI
        # in 3D, in which case the ==0 assertion may become imprecise and we would need a tolerance to compare with a
        # double value.
        n_y_pixels = int(self.roi_data['Y stop'][0] - self.roi_data['Y start'][0])
        n_x_pixels = int(self.roi_data['X stop'][0] - self.roi_data['X start'][0])
        if self.roi_data['Angle (deg)'][0] == 0:
            self.n_lines_per_roi = n_y_pixels
            self.n_pixels_per_line = n_x_pixels
        else:
            self.n_lines_per_roi = n_x_pixels
            self.n_pixels_per_line = n_y_pixels
        if self.n_pixels_per_line == 0 and self.n_lines_per_roi == 0:
            # assume we are in pointing mode in this case, probably better if this were passed as an argument?
            # this class would need to know about the pointing mode then though.
            self.n_pixels_per_line = 1
            self.n_lines_per_roi = 1


class LabViewTimingsPre2018(LabViewTimings):

    def __init__(self, relative_times_path, roi_path, dwell_time):
        super().__init__(relative_times_path, roi_path, dwell_time)
        self._format_pixel_time_offsets()

    def _read_relative_times_file(self, file_path):
        raw_data = pd.read_csv(file_path, names=('RelativeTime', 'CycleTime'),
                               sep='\t', dtype=np.float64) / 1e6  # convert to seconds
        self.pixel_time_offsets = raw_data['RelativeTime']
        self.cycle_time = raw_data['CycleTime'][0]

    def _format_pixel_time_offsets(self):
        row_increments = np.arange(self.n_pixels_per_line) * self.dwell_time
        pixel_time_offsets_by_roi = {}
        for roi_index in np.arange(self.n_rois):
            start_index = self.n_lines_per_roi * roi_index
            end_index = start_index + self.n_lines_per_roi
            row_offsets = self.pixel_time_offsets[start_index:end_index]
            pixel_time_offsets_by_roi[roi_index] = row_offsets[:, np.newaxis] + row_increments
        self.pixel_time_offsets = pixel_time_offsets_by_roi


class LabViewTimings231(LabViewTimings):

    def __init__(self, relative_times_path, roi_path, dwell_time, n_cycles_per_trial, n_trials):
        super().__init__(relative_times_path, roi_path, dwell_time)
        self._format_pixel_time_offsets(n_cycles_per_trial, n_trials)

    def _read_relative_times_file(self, file_path):
        raw_data = pd.read_csv(file_path, sep='\t', dtype=np.float64) / 1e6
        self.pixel_time_offsets = raw_data['Image Time [us]']
        self.pixel_time_offsets = self.pixel_time_offsets[self.pixel_time_offsets != 0]

    def _format_pixel_time_offsets(self, n_cycles_per_trial, n_trials):
        pixel_time_offsets_by_roi = {}
        n_lines_per_cycle = self.n_rois * self.n_lines_per_roi
        row_increments = np.arange(self.n_pixels_per_line) * self.dwell_time
        for roi_index in np.arange(0, self.n_rois):
            pixel_time_offsets_by_roi[roi_index] = []
            for cycle_index in np.arange(0, n_cycles_per_trial * n_trials):
                start_index = self.n_lines_per_roi * roi_index + n_lines_per_cycle * cycle_index
                end_index = start_index + self.n_lines_per_roi
                pixel_time_offsets_by_roi[roi_index].append(self.pixel_time_offsets.values[start_index:end_index])
            pixel_time_offsets_by_roi[roi_index] = np.reshape(pixel_time_offsets_by_roi[roi_index],
                                                              (n_trials * n_cycles_per_trial, self.n_lines_per_roi))
            pixel_time_offsets_by_roi[roi_index] = pixel_time_offsets_by_roi[roi_index][:, :, np.newaxis] + row_increments

        # estimate time for one cycle by averaging the time it takes for the first cycle of each trial.
        # The n_pixels_per_line * pixel_dwell_time contribution of the last line is negligible.
        first_cycle_times_for_each_trial = []
        for trial_index in list(range(n_trials)):
            first_cycle_times_for_each_trial.append(pixel_time_offsets_by_roi[self.n_rois - 1]
                                                    [trial_index * n_cycles_per_trial]
                                                    [self.n_lines_per_roi - 1][0])

        self.cycle_time = np.mean(first_cycle_times_for_each_trial)
        self.pixel_time_offsets = pixel_time_offsets_by_roi
